from __future__ import annotations

import datetime as dt
from typing import Any
from typing import cast

import numpy as np
from hypothesis import assume
from hypothesis.errors import InvalidArgument
from hypothesis.extra.numpy import array_shapes
from hypothesis.extra.numpy import arrays
from hypothesis.strategies import DrawFn
from hypothesis.strategies import SearchStrategy
from hypothesis.strategies import booleans
from hypothesis.strategies import composite
from hypothesis.strategies import integers
from hypothesis.strategies import none
from hypothesis.strategies import nothing
from hypothesis.strategies import sampled_from
from numpy import concatenate
from numpy import datetime64
from numpy import expand_dims
from numpy import iinfo
from numpy import int32
from numpy import int64
from numpy import uint32
from numpy import uint64
from numpy import zeros
from numpy.typing import NDArray

from utilities.math import IntNonNeg
from utilities.numpy import DATE_MAX_AS_INT
from utilities.numpy import DATE_MIN_AS_INT
from utilities.numpy import DATETIME_MAX_AS_INT
from utilities.numpy import DATETIME_MIN_AS_INT
from utilities.numpy import Datetime64Kind
from utilities.numpy import Datetime64Unit
from utilities.numpy import EmptyNumpyConcatenateError
from utilities.numpy import NDArrayB
from utilities.numpy import NDArrayD
from utilities.numpy import NDArrayD1
from utilities.numpy import NDArrayDD1
from utilities.numpy import NDArrayDus1
from utilities.numpy import NDArrayF
from utilities.numpy import NDArrayI
from utilities.numpy import NDArrayO
from utilities.numpy import date_to_datetime64
from utilities.numpy import datetime64_to_int
from utilities.numpy import datetime64_unit_to_dtype
from utilities.numpy import datetime64_unit_to_kind
from utilities.numpy import datetime_to_datetime64
from utilities.numpy import redirect_to_empty_numpy_concatenate_error

from .hypothesis import MaybeSearchStrategy  # noqa: TID252
from .hypothesis import Shape  # noqa: TID252
from .hypothesis import floats_extra  # noqa: TID252
from .hypothesis import lift_draw  # noqa: TID252
from .hypothesis import lists_fixed_length  # noqa: TID252
from .hypothesis import text_ascii  # noqa: TID252


@composite
def bool_arrays(
    _draw: DrawFn,
    /,
    *,
    shape: MaybeSearchStrategy[Shape] = array_shapes(),
    fill: SearchStrategy[Any] | None = None,
    unique: MaybeSearchStrategy[bool] = False,
) -> NDArrayB:
    """Strategy for generating arrays of booleans."""
    draw = lift_draw(_draw)
    strategy = cast(
        SearchStrategy[NDArrayB],
        arrays(
            bool,
            draw(shape),
            elements=booleans(),
            fill=fill,
            unique=draw(unique),
        ),
    )
    return draw(strategy)


@composite
def concatenated_arrays(
    _draw: DrawFn,
    strategy: SearchStrategy[NDArray[Any]],
    size: MaybeSearchStrategy[IntNonNeg],
    fallback: Shape,
    /,
    *,
    dtype: Any = float,
) -> NDArray[Any]:
    """Strategy for generating arrays from lower-dimensional strategies."""
    draw = lift_draw(_draw)
    size_ = draw(size)
    arrays = draw(lists_fixed_length(strategy, size_))
    expanded = [expand_dims(array, axis=0) for array in arrays]
    try:
        return concatenate(expanded)
    except ValueError as error:
        try:
            redirect_to_empty_numpy_concatenate_error(error)
        except EmptyNumpyConcatenateError:
            if isinstance(fallback, int):
                shape = size_, fallback
            else:
                shape = (size_, *fallback)
            return zeros(shape, dtype=dtype)


@composite
def datetime64_dtypes(
    _draw: DrawFn, /, *, kind: MaybeSearchStrategy[Datetime64Kind | None] = None
) -> Any:
    """Strategy for generating datetime64 dtypes."""
    draw = lift_draw(_draw)
    unit = draw(datetime64_units(kind=kind))
    return datetime64_unit_to_dtype(unit)


def datetime64_kinds() -> SearchStrategy[Datetime64Kind]:
    """Strategy for generating datetime64 kinds."""
    kinds: list[Datetime64Kind] = ["date", "time"]
    return sampled_from(kinds)


@composite
def datetime64_units(
    _draw: DrawFn, /, *, kind: MaybeSearchStrategy[Datetime64Kind | None] = None
) -> Datetime64Unit:
    """Strategy for generating datetime64 units."""
    draw = lift_draw(_draw)
    units: list[Datetime64Unit] = [
        "Y",
        "M",
        "W",
        "D",
        "h",
        "m",
        "s",
        "ms",
        "us",
        "ns",
        "ps",
        "fs",
        "as",
    ]
    kind_ = draw(kind)
    if kind_ is not None:
        units = [
            unit for unit in units if datetime64_unit_to_kind(unit) == kind_
        ]
    return draw(sampled_from(units))


@composite
def datetime64_arrays(
    _draw: DrawFn,
    /,
    *,
    shape: MaybeSearchStrategy[Shape] = array_shapes(),
    unit: MaybeSearchStrategy[Datetime64Unit | None] = None,
    min_value: MaybeSearchStrategy[datetime64 | int | dt.date | None] = None,
    max_value: MaybeSearchStrategy[datetime64 | int | dt.date | None] = None,
    valid_dates: MaybeSearchStrategy[bool] = False,
    valid_datetimes: MaybeSearchStrategy[bool] = False,
    fill: SearchStrategy[Any] | None = None,
    unique: MaybeSearchStrategy[bool] = False,
) -> NDArrayD:
    """Strategy for generating arrays of datetime64s."""
    draw = lift_draw(_draw)
    if (unit_ := draw(unit)) is None:
        unit_ = draw(datetime64_units())
    dtype = datetime64_unit_to_dtype(cast(Datetime64Unit, unit_))
    elements = datetime64s(
        unit=cast(Datetime64Unit, unit_),
        min_value=min_value,
        max_value=max_value,
        valid_dates=valid_dates,
        valid_datetimes=valid_datetimes,
    )
    return draw(
        arrays(
            dtype,
            draw(shape),
            elements=elements,
            fill=fill,
            unique=draw(unique),
        )
    )


@composite
def datetime64_indexes(
    _draw: DrawFn,
    /,
    *,
    n: MaybeSearchStrategy[IntNonNeg] = integers(0, 10),
    unit: MaybeSearchStrategy[Datetime64Unit | None] = None,
    min_value: MaybeSearchStrategy[datetime64 | int | dt.date | None] = None,
    max_value: MaybeSearchStrategy[datetime64 | int | dt.date | None] = None,
    valid_dates: MaybeSearchStrategy[bool] = False,
    valid_datetimes: MaybeSearchStrategy[bool] = False,
    unique: MaybeSearchStrategy[bool] = True,
    sort: MaybeSearchStrategy[bool] = True,
) -> NDArrayD1:
    """Strategy for generating indexes of datetime64s."""
    draw = lift_draw(_draw)
    array = draw(
        datetime64_arrays(
            shape=n,
            unit=unit,
            min_value=min_value,
            max_value=max_value,
            valid_dates=valid_dates,
            valid_datetimes=valid_datetimes,
            fill=nothing(),
            unique=unique,
        )
    )
    return np.sort(array) if sort else array


def datetime64D_indexes(  # noqa: N802
    *,
    n: MaybeSearchStrategy[IntNonNeg] = integers(0, 10),
    min_value: MaybeSearchStrategy[datetime64 | int | dt.date | None] = None,
    max_value: MaybeSearchStrategy[datetime64 | int | dt.date | None] = None,
    valid_dates: MaybeSearchStrategy[bool] = True,
    unique: MaybeSearchStrategy[bool] = True,
    sort: MaybeSearchStrategy[bool] = True,
) -> SearchStrategy[NDArrayDD1]:
    """Strategy for generating arrays of dates."""
    return datetime64_indexes(
        n=n,
        unit="D",
        min_value=min_value,
        max_value=max_value,
        valid_dates=valid_dates,
        unique=unique,
        sort=sort,
    )


@composite
def datetime64s(
    _draw: DrawFn,
    /,
    *,
    unit: MaybeSearchStrategy[Datetime64Unit | None] = None,
    min_value: MaybeSearchStrategy[datetime64 | int | dt.date | None] = None,
    max_value: MaybeSearchStrategy[datetime64 | int | dt.date | None] = None,
    valid_dates: MaybeSearchStrategy[bool] = False,
    valid_datetimes: MaybeSearchStrategy[bool] = False,
) -> datetime64:
    """Strategy for generating datetime64s."""
    draw = lift_draw(_draw)
    unit_ = draw(unit)
    min_value_, max_value_ = (
        _datetime64s_convert(draw(mv)) for mv in (min_value, max_value)
    )
    valid_dates_, valid_datetimes_ = (
        draw(vd) for vd in (valid_dates, valid_datetimes)
    )
    if valid_dates_:
        unit_, min_value_, max_value_ = _datetime64s_check_valid_dates(
            unit=cast(Datetime64Unit | None, unit_),
            min_value=min_value_,
            max_value=max_value_,
        )
    if valid_datetimes_:
        unit_, min_value_, max_value_ = _datetime64s_check_valid_datetimes(
            unit=cast(Datetime64Unit | None, unit_),
            min_value=min_value_,
            max_value=max_value_,
        )
    i = draw(int64s(min_value=min_value_, max_value=max_value_))
    _ = assume(i != iinfo(int64).min)
    if unit_ is None:
        unit_ = draw(datetime64_units())
    return datetime64(i, unit_)


def _datetime64s_convert(
    value: int | datetime64 | dt.date | None, /
) -> int | None:
    """Convert a min/max value supplied into `datetime64s`."""
    if (value is None) or isinstance(value, int):
        return value
    if isinstance(value, datetime64):
        return datetime64_to_int(value)
    if isinstance(value, dt.datetime):
        return _datetime64s_convert(datetime_to_datetime64(value))
    return _datetime64s_convert(date_to_datetime64(value))


def _datetime64s_check_valid_dates(
    *,
    unit: Datetime64Unit | None = None,
    min_value: int | None = None,
    max_value: int | None = None,
) -> tuple[Datetime64Unit, int | None, int | None]:
    """Check/clip the bounds to generate valid `dt.date`s."""
    if (unit is not None) and (unit != "D"):
        msg = f"{unit=}"
        raise InvalidArgument(msg)
    if min_value is None:
        min_value = DATE_MIN_AS_INT
    else:
        min_value = max(min_value, DATE_MIN_AS_INT)
    if max_value is None:
        max_value = DATE_MAX_AS_INT
    else:
        max_value = min(max_value, DATE_MAX_AS_INT)
    return "D", min_value, max_value


def _datetime64s_check_valid_datetimes(
    *,
    unit: Datetime64Unit | None = None,
    min_value: int | None = None,
    max_value: int | None = None,
) -> tuple[Datetime64Unit, int | None, int | None]:
    """Check/clip the bounds to generate valid `dt.datetime`s."""
    if (unit is not None) and (unit != "us"):
        msg = f"{unit=}"
        raise InvalidArgument(msg)
    if min_value is None:
        min_value = DATETIME_MIN_AS_INT
    else:
        min_value = max(min_value, DATETIME_MIN_AS_INT)
    if max_value is None:
        max_value = DATETIME_MAX_AS_INT
    else:
        max_value = min(max_value, DATETIME_MAX_AS_INT)
    return "us", min_value, max_value


def datetime64us_indexes(
    *,
    n: MaybeSearchStrategy[IntNonNeg] = integers(0, 10),
    min_value: MaybeSearchStrategy[
        datetime64 | int | dt.datetime | None
    ] = None,
    max_value: MaybeSearchStrategy[
        datetime64 | int | dt.datetime | None
    ] = None,
    valid_datetimes: MaybeSearchStrategy[bool] = True,
    unique: MaybeSearchStrategy[bool] = True,
    sort: MaybeSearchStrategy[bool] = True,
) -> SearchStrategy[NDArrayDus1]:
    """Strategy for generating arrays of datetimes."""
    return datetime64_indexes(
        n=n,
        unit="us",
        min_value=min_value,
        max_value=max_value,
        valid_datetimes=valid_datetimes,
        unique=unique,
        sort=sort,
    )


@composite
def float_arrays(
    _draw: DrawFn,
    /,
    *,
    shape: MaybeSearchStrategy[Shape] = array_shapes(),
    min_value: MaybeSearchStrategy[float | None] = None,
    max_value: MaybeSearchStrategy[float | None] = None,
    allow_nan: MaybeSearchStrategy[bool] = False,
    allow_inf: MaybeSearchStrategy[bool] = False,
    allow_pos_inf: MaybeSearchStrategy[bool] = False,
    allow_neg_inf: MaybeSearchStrategy[bool] = False,
    integral: MaybeSearchStrategy[bool] = False,
    fill: SearchStrategy[Any] | None = None,
    unique: MaybeSearchStrategy[bool] = False,
) -> NDArrayF:
    """Strategy for generating arrays of floats."""
    draw = lift_draw(_draw)
    elements = floats_extra(
        min_value=min_value,
        max_value=max_value,
        allow_nan=allow_nan,
        allow_inf=allow_inf,
        allow_pos_inf=allow_pos_inf,
        allow_neg_inf=allow_neg_inf,
        integral=integral,
    )
    strategy = cast(
        SearchStrategy[NDArrayF],
        arrays(
            float,
            draw(shape),
            elements=elements,
            fill=fill,
            unique=draw(unique),
        ),
    )
    return draw(strategy)


@composite
def int_arrays(
    _draw: DrawFn,
    /,
    *,
    shape: MaybeSearchStrategy[Shape] = array_shapes(),
    min_value: MaybeSearchStrategy[int | None] = None,
    max_value: MaybeSearchStrategy[int | None] = None,
    fill: SearchStrategy[Any] | None = None,
    unique: MaybeSearchStrategy[bool] = False,
) -> NDArrayI:
    """Strategy for generating arrays of ints."""
    draw = lift_draw(_draw)
    info = iinfo(int64)
    min_value_, max_value_ = draw(min_value), draw(max_value)
    min_value_use = info.min if min_value_ is None else min_value_
    max_value_use = info.max if max_value_ is None else max_value_
    elements = integers(min_value=min_value_use, max_value=max_value_use)
    strategy = cast(
        SearchStrategy[NDArrayI],
        arrays(
            int, draw(shape), elements=elements, fill=fill, unique=draw(unique)
        ),
    )
    return draw(strategy)


def int32s(
    *,
    min_value: MaybeSearchStrategy[int | None] = None,
    max_value: MaybeSearchStrategy[int | None] = None,
) -> SearchStrategy[int]:
    """Strategy for generating int32s."""
    return _fixed_width_ints(int32, min_value=min_value, max_value=max_value)


def int64s(
    *,
    min_value: MaybeSearchStrategy[int | None] = None,
    max_value: MaybeSearchStrategy[int | None] = None,
) -> SearchStrategy[int]:
    """Strategy for generating int64s."""
    return _fixed_width_ints(int64, min_value=min_value, max_value=max_value)


@composite
def str_arrays(
    _draw: DrawFn,
    /,
    *,
    shape: MaybeSearchStrategy[Shape] = array_shapes(),
    min_size: MaybeSearchStrategy[int] = 0,
    max_size: MaybeSearchStrategy[int | None] = None,
    allow_none: MaybeSearchStrategy[bool] = False,
    fill: SearchStrategy[Any] | None = None,
    unique: MaybeSearchStrategy[bool] = False,
) -> NDArrayO:
    """Strategy for generating arrays of strings."""
    draw = lift_draw(_draw)
    elements = text_ascii(min_size=min_size, max_size=max_size)
    if draw(allow_none):
        elements |= none()
    strategy = cast(
        SearchStrategy[NDArrayO],
        arrays(
            object,
            draw(shape),
            elements=elements,
            fill=fill,
            unique=draw(unique),
        ),
    )
    return draw(strategy)


def uint32s(
    *,
    min_value: MaybeSearchStrategy[int | None] = None,
    max_value: MaybeSearchStrategy[int | None] = None,
) -> SearchStrategy[int]:
    """Strategy for generating uint32s."""
    return _fixed_width_ints(uint32, min_value=min_value, max_value=max_value)


def uint64s(
    *,
    min_value: MaybeSearchStrategy[int | None] = None,
    max_value: MaybeSearchStrategy[int | None] = None,
) -> SearchStrategy[int]:
    """Strategy for generating uint64s."""
    return _fixed_width_ints(uint64, min_value=min_value, max_value=max_value)


@composite
def _fixed_width_ints(
    _draw: DrawFn,
    dtype: Any,
    /,
    *,
    min_value: MaybeSearchStrategy[int | None] = None,
    max_value: MaybeSearchStrategy[int | None] = None,
) -> int:
    """Strategy for generating int64s."""
    draw = lift_draw(_draw)
    min_value_, max_value_ = (draw(mv) for mv in (min_value, max_value))
    info = iinfo(dtype)
    min_value_ = info.min if min_value_ is None else max(min_value_, info.min)
    max_value_use = (
        info.max if max_value_ is None else min(info.max, max_value_)
    )
    return draw(integers(min_value_, max_value_use))



from __future__ import annotations

from math import isclose
from math import isfinite
from math import isnan
from typing import Annotated

# functions


def is_at_least(
    x: float,
    y: float,
    /,
    *,
    rel_tol: float | None = None,
    abs_tol: float | None = None,
) -> bool:
    """Check if -inf < x < inf and x == int(x)."""
    return (x >= y) or _is_close(x, y, rel_tol=rel_tol, abs_tol=abs_tol)


def is_at_least_or_nan(
    x: float,
    y: float,
    /,
    *,
    rel_tol: float | None = None,
    abs_tol: float | None = None,
) -> bool:
    """Check if x >= y or x == nan."""
    return is_at_least(x, y, rel_tol=rel_tol, abs_tol=abs_tol) or isnan(x)


def is_at_most(
    x: float,
    y: float,
    /,
    *,
    rel_tol: float | None = None,
    abs_tol: float | None = None,
) -> bool:
    """Check if x <= y."""
    return (x <= y) or _is_close(x, y, rel_tol=rel_tol, abs_tol=abs_tol)


def is_at_most_or_nan(
    x: float,
    y: float,
    /,
    *,
    rel_tol: float | None = None,
    abs_tol: float | None = None,
) -> bool:
    """Check if x <= y or x == nan."""
    return is_at_most(x, y, rel_tol=rel_tol, abs_tol=abs_tol) or isnan(x)


def is_between(
    x: float,
    low: float,
    high: float,
    /,
    *,
    rel_tol: float | None = None,
    abs_tol: float | None = None,
) -> bool:
    """Check if low <= x <= high."""
    return is_at_least(x, low, rel_tol=rel_tol, abs_tol=abs_tol) and is_at_most(
        x, high, rel_tol=rel_tol, abs_tol=abs_tol
    )


def is_between_or_nan(
    x: float,
    low: float,
    high: float,
    /,
    *,
    rel_tol: float | None = None,
    abs_tol: float | None = None,
) -> bool:
    """Check if low <= x <= high or x == nan."""
    return is_between(x, low, high, rel_tol=rel_tol, abs_tol=abs_tol) or isnan(
        x
    )


def is_finite_and_integral(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if -inf < x < inf and x == int(x)."""
    return isfinite(x) & is_integral(x, rel_tol=rel_tol, abs_tol=abs_tol)


def is_finite_and_integral_or_nan(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if -inf < x < inf and x == int(x), or x == nan."""
    return is_finite_and_integral(x, rel_tol=rel_tol, abs_tol=abs_tol) | isnan(
        x
    )


def is_finite_and_negative(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if -inf < x < 0."""
    return isfinite(x) and is_negative(x, rel_tol=rel_tol, abs_tol=abs_tol)


def is_finite_and_negative_or_nan(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if -inf < x < 0 or x == nan."""
    return is_finite_and_negative(x, rel_tol=rel_tol, abs_tol=abs_tol) or isnan(
        x
    )


def is_finite_and_non_negative(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if 0 <= x < inf."""
    return isfinite(x) and is_non_negative(x, rel_tol=rel_tol, abs_tol=abs_tol)


def is_finite_and_non_negative_or_nan(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if 0 <= x < inf or x == nan."""
    return is_finite_and_non_negative(
        x, rel_tol=rel_tol, abs_tol=abs_tol
    ) or isnan(x)


def is_finite_and_non_positive(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if -inf < x <= 0."""
    return isfinite(x) and is_non_positive(x, rel_tol=rel_tol, abs_tol=abs_tol)


def is_finite_and_non_positive_or_nan(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if -inf < x <= 0 or x == nan."""
    return is_finite_and_non_positive(
        x, rel_tol=rel_tol, abs_tol=abs_tol
    ) or isnan(x)


def is_finite_and_non_zero(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if -inf < x < inf, x != 0."""
    return isfinite(x) and is_non_zero(x, rel_tol=rel_tol, abs_tol=abs_tol)


def is_finite_and_non_zero_or_nan(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if x != 0 or x == nan."""
    return is_finite_and_non_zero(x, rel_tol=rel_tol, abs_tol=abs_tol) or isnan(
        x
    )


def is_finite_and_positive(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if 0 < x < inf."""
    return isfinite(x) and is_positive(x, rel_tol=rel_tol, abs_tol=abs_tol)


def is_finite_and_positive_or_nan(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if 0 < x < inf or x == nan."""
    return is_finite_and_positive(x, rel_tol=rel_tol, abs_tol=abs_tol) or isnan(
        x
    )


def is_finite_or_nan(x: float, /) -> bool:
    """Check if -inf < x < inf or x == nan."""
    return isfinite(x) or isnan(x)


def is_greater_than(
    x: float,
    y: float,
    /,
    *,
    rel_tol: float | None = None,
    abs_tol: float | None = None,
) -> bool:
    """Check if x > y."""
    return (x > y) and not _is_close(x, y, rel_tol=rel_tol, abs_tol=abs_tol)


def is_greater_than_or_nan(
    x: float,
    y: float,
    /,
    *,
    rel_tol: float | None = None,
    abs_tol: float | None = None,
) -> bool:
    """Check if x > y or x == nan."""
    return is_greater_than(x, y, rel_tol=rel_tol, abs_tol=abs_tol) or isnan(x)


def is_integral(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if x == int(x)."""
    try:
        rounded = round(x)
    except (OverflowError, ValueError):
        rounded = x
    return _is_close(x, rounded, rel_tol=rel_tol, abs_tol=abs_tol)


def is_integral_or_nan(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if x == int(x) or x == nan."""
    return is_integral(x, rel_tol=rel_tol, abs_tol=abs_tol) | isnan(x)


def is_less_than(
    x: float,
    y: float,
    /,
    *,
    rel_tol: float | None = None,
    abs_tol: float | None = None,
) -> bool:
    """Check if x < y."""
    return (x < y) and not _is_close(x, y, rel_tol=rel_tol, abs_tol=abs_tol)


def is_less_than_or_nan(
    x: float,
    y: float,
    /,
    *,
    rel_tol: float | None = None,
    abs_tol: float | None = None,
) -> bool:
    """Check if x < y or x == nan."""
    return is_less_than(x, y, rel_tol=rel_tol, abs_tol=abs_tol) or isnan(x)


def is_negative(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if x < 0."""
    return is_less_than(x, 0.0, rel_tol=rel_tol, abs_tol=abs_tol)


def is_negative_or_nan(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if x < 0 or x == nan."""
    return is_negative(x, rel_tol=rel_tol, abs_tol=abs_tol) or isnan(x)


def is_non_negative(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if x >= 0."""
    return is_at_least(x, 0.0, rel_tol=rel_tol, abs_tol=abs_tol)


def is_non_negative_or_nan(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if x >= 0 or x == nan."""
    return is_non_negative(x, rel_tol=rel_tol, abs_tol=abs_tol) or isnan(x)


def is_non_positive(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if x <= 0."""
    return is_at_most(x, 0.0, rel_tol=rel_tol, abs_tol=abs_tol)


def is_non_positive_or_nan(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if x <=0 or x == nan."""
    return is_non_positive(x, rel_tol=rel_tol, abs_tol=abs_tol) or isnan(x)


def is_non_zero(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if x != 0."""
    return not _is_close(x, 0.0, rel_tol=rel_tol, abs_tol=abs_tol)


def is_non_zero_or_nan(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if x != 0 or x == nan."""
    return is_non_zero(x, rel_tol=rel_tol, abs_tol=abs_tol) or isnan(x)


def is_positive(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if x > 0."""
    return is_greater_than(x, 0, rel_tol=rel_tol, abs_tol=abs_tol)


def is_positive_or_nan(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if x > 0 or x == nan."""
    return is_positive(x, rel_tol=rel_tol, abs_tol=abs_tol) or isnan(x)


def is_zero(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if x == 0."""
    return _is_close(x, 0.0, rel_tol=rel_tol, abs_tol=abs_tol)


def is_zero_or_finite_and_non_micro(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if x == 0, or -inf < x < inf and ~isclose(x, 0)."""
    zero = 0.0
    return (x == zero) or is_finite_and_non_zero(
        x, rel_tol=rel_tol, abs_tol=abs_tol
    )


def is_zero_or_finite_and_non_micro_or_nan(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if x == 0, or -inf < x < inf and ~isclose(x, 0), or x == nan."""
    return is_zero_or_finite_and_non_micro(
        x, rel_tol=rel_tol, abs_tol=abs_tol
    ) or isnan(x)


def is_zero_or_nan(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if x > 0 or x == nan."""
    return is_zero(x, rel_tol=rel_tol, abs_tol=abs_tol) or isnan(x)


def is_zero_or_non_micro(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if x == 0 or ~isclose(x, 0)."""
    zero = 0.0
    return (x == zero) or is_non_zero(x, rel_tol=rel_tol, abs_tol=abs_tol)


def is_zero_or_non_micro_or_nan(
    x: float, /, *, rel_tol: float | None = None, abs_tol: float | None = None
) -> bool:
    """Check if x == 0 or ~isclose(x, 0) or x == nan."""
    return is_zero_or_non_micro(x, rel_tol=rel_tol, abs_tol=abs_tol) or isnan(x)


def _is_close(
    x: float,
    y: float,
    /,
    *,
    rel_tol: float | None = None,
    abs_tol: float | None = None,
) -> bool:
    """Check if x == y."""
    return isclose(
        x,
        y,
        **({} if rel_tol is None else {"rel_tol": rel_tol}),
        **({} if abs_tol is None else {"abs_tol": abs_tol}),
    )


# annotations


IntNeg = Annotated[int, is_negative]
IntNonNeg = Annotated[int, is_non_negative]
IntNonPos = Annotated[int, is_non_positive]
IntNonZr = Annotated[int, is_non_zero]
IntPos = Annotated[int, is_positive]
IntZr = Annotated[int, is_zero]

# float
FloatFin = Annotated[float, isfinite]
FloatFinInt = Annotated[float, is_finite_and_integral]
FloatFinIntNan = Annotated[float, is_finite_and_integral_or_nan]
FloatFinNeg = Annotated[float, is_finite_and_negative]
FloatFinNegNan = Annotated[float, is_finite_and_negative_or_nan]
FloatFinNonNeg = Annotated[float, is_finite_and_non_negative]
FloatFinNonNegNan = Annotated[float, is_finite_and_non_negative_or_nan]
FloatFinNonPos = Annotated[float, is_finite_and_non_positive]
FloatFinNonPosNan = Annotated[float, is_finite_and_non_positive_or_nan]
FloatFinNonZr = Annotated[float, is_finite_and_non_zero]
FloatFinNonZrNan = Annotated[float, is_finite_and_non_zero_or_nan]
FloatFinPos = Annotated[float, is_finite_and_positive]
FloatFinPosNan = Annotated[float, is_finite_and_positive_or_nan]
FloatFinNan = Annotated[float, is_finite_or_nan]
FloatInt = Annotated[float, is_integral]
FloatIntNan = Annotated[float, is_integral_or_nan]
FloatNeg = Annotated[float, is_negative]
FloatNegNan = Annotated[float, is_negative_or_nan]
FloatNonNeg = Annotated[float, is_non_negative]
FloatNonNegNan = Annotated[float, is_non_negative_or_nan]
FloatNonPos = Annotated[float, is_non_positive]
FloatNonPosNan = Annotated[float, is_non_positive_or_nan]
FloatNonZr = Annotated[float, is_non_zero]
FloatNonZrNan = Annotated[float, is_non_zero_or_nan]
FloatPos = Annotated[float, is_positive]
FloatPosNan = Annotated[float, is_positive_or_nan]
FloatZr = Annotated[float, is_zero]
FloatZrFinNonMic = Annotated[float, is_zero_or_finite_and_non_micro]
FloatZrFinNonMicNan = Annotated[float, is_zero_or_finite_and_non_micro_or_nan]
FloatZrNan = Annotated[float, is_zero_or_nan]
FloatZrNonMic = Annotated[float, is_zero_or_non_micro]
FloatZrNonMicNan = Annotated[float, is_zero_or_non_micro_or_nan]



from __future__ import annotations

import datetime as dt
from enum import Enum
from enum import auto
from functools import partial
from pathlib import Path
from typing import Any
from typing import Literal
from typing import cast

from freezegun import freeze_time
from hypothesis import assume
from hypothesis import given
from hypothesis import settings
from hypothesis.strategies import DataObject
from hypothesis.strategies import booleans
from hypothesis.strategies import data
from hypothesis.strategies import dates
from hypothesis.strategies import sampled_from
from hypothesis.strategies import times
from luigi import BoolParameter
from luigi import Task
from luigi.notifications import smtp
from luigi.task import Parameter
from pytest import mark
from pytest import param
from typing_extensions import override

from utilities.datetime import serialize_date
from utilities.datetime import serialize_datetime
from utilities.datetime import serialize_time
from utilities.hypothesis import datetimes_utc
from utilities.hypothesis import namespace_mixins
from utilities.hypothesis import temp_paths
from utilities.luigi import AwaitTask
from utilities.luigi import AwaitTime
from utilities.luigi import DateHourParameter
from utilities.luigi import DateMinuteParameter
from utilities.luigi import DateParameter
from utilities.luigi import DateSecondParameter
from utilities.luigi import EnumParameter
from utilities.luigi import ExternalFile
from utilities.luigi import ExternalTask
from utilities.luigi import PathTarget
from utilities.luigi import TimeParameter
from utilities.luigi import WeekdayParameter
from utilities.luigi import build
from utilities.luigi import clone
from utilities.luigi import get_dependencies_downstream
from utilities.luigi import get_dependencies_upstream
from utilities.luigi import get_task_classes
from utilities.luigi.luigi import _yield_task_classes


class TestAwaitTask:
    @given(namespace_mixin=namespace_mixins(), is_complete=booleans())
    def test_main(self, *, namespace_mixin: Any, is_complete: bool) -> None:
        class Example(namespace_mixin, Task):
            is_complete = cast(bool, BoolParameter())

            @override
            def complete(self) -> bool:
                return self.is_complete

        example = Example(is_complete=is_complete)
        task: AwaitTask[Any] = cast(Any, AwaitTask)(example)
        result = task.complete()
        assert result is is_complete


class TestAwaitTime:
    @given(time_start=datetimes_utc(), time_now=datetimes_utc())
    def test_main(self, time_start: dt.datetime, time_now: dt.datetime) -> None:
        _ = assume(time_start.microsecond == 0)
        task: AwaitTime = cast(Any, AwaitTime)(time_start)
        with freeze_time(time_now):
            result = task.exists()
        expected = time_now >= time_start
        assert result is expected


class TestBuild:
    @given(namespace_mixin=namespace_mixins())
    def test_main(self, namespace_mixin: Any) -> None:
        class Example(namespace_mixin, Task):
            ...

        _ = build([Example()], local_scheduler=True)


class TestClone:
    @given(namespace_mixin=namespace_mixins(), truth=booleans())
    def test_main(self, *, namespace_mixin: Any, truth: bool) -> None:
        class A(namespace_mixin, Task):
            truth = cast(bool, BoolParameter())

        class B(namespace_mixin, Task):
            truth = cast(bool, BoolParameter())

        a = A(truth)
        result = clone(a, B)
        expected = B(truth)
        assert result is expected

    @given(namespace_mixin=namespace_mixins(), truth=booleans())
    def test_await(self, *, namespace_mixin: Any, truth: bool) -> None:
        class A(namespace_mixin, Task):
            truth = cast(bool, BoolParameter())

        class B(namespace_mixin, Task):
            truth = cast(bool, BoolParameter())

        a = A(truth)
        result = clone(a, B, await_=True)
        expected = AwaitTask(B(truth))
        assert result is expected


class TestDateParameter:
    @given(data=data(), date=dates())
    def test_main(self, data: DataObject, date: dt.date) -> None:
        param = DateParameter()
        input_ = data.draw(sampled_from([date, serialize_date(date)]))
        norm = param.normalize(input_)
        assert param.parse(param.serialize(norm)) == norm


class TestDateTimeParameter:
    @given(data=data(), datetime=datetimes_utc())
    @mark.parametrize(
        "param_cls",
        [
            param(DateHourParameter),
            param(DateMinuteParameter),
            param(DateSecondParameter),
        ],
    )
    def test_main(
        self,
        data: DataObject,
        datetime: dt.datetime,
        param_cls: type[Parameter],
    ) -> None:
        param = param_cls()
        input_ = data.draw(
            sampled_from([datetime, serialize_datetime(datetime)])
        )
        norm = param.normalize(input_)
        assert param.parse(param.serialize(norm)) == norm


class TestEnumParameter:
    @given(data=data())
    def test_main(self, data: DataObject) -> None:
        class Example(Enum):
            member = auto()

        param = EnumParameter(Example)
        input_ = data.draw(sampled_from([Example.member, "member"]))
        norm = param.normalize(input_)
        assert param.parse(param.serialize(norm)) == norm


class TestExternalFile:
    @given(namespace_mixin=namespace_mixins(), root=temp_paths())
    def test_main(self, namespace_mixin: Any, root: Path) -> None:
        path = root.joinpath("file")

        class Example(namespace_mixin, ExternalFile):
            ...

        task = Example(path)
        assert not task.exists()
        path.touch()
        assert task.exists()


class TestExternalTask:
    @given(namespace_mixin=namespace_mixins(), is_complete=booleans())
    def test_main(self, *, namespace_mixin: Any, is_complete: bool) -> None:
        class Example(namespace_mixin, ExternalTask):
            is_complete = cast(bool, BoolParameter())

            @override
            def exists(self) -> bool:
                return self.is_complete

        task = Example(is_complete=is_complete)
        result = task.exists()
        assert result is is_complete


class TestGetDependencies:
    @given(namespace_mixin=namespace_mixins())
    @settings(max_examples=1)
    def test_main(self, namespace_mixin: Any) -> None:
        class A(namespace_mixin, Task):
            ...

        class B(namespace_mixin, Task):
            @override
            def requires(self) -> A:
                return clone(self, A)

        class C(namespace_mixin, Task):
            @override
            def requires(self) -> B:
                return clone(self, B)

        a, b, c = A(), B(), C()
        ((up_a, down_a), (up_b, down_b), (up_c, down_c)) = map(
            self._get_sets, [a, b, c]
        )
        assert up_a == set()
        assert down_a == {b}
        assert up_b == {a}
        assert down_b == {c}
        assert up_c == {b}
        assert down_c == set()

        (
            (up_a_rec, down_a_rec),
            (up_b_rec, down_b_rec),
            (up_c_rec, down_c_rec),
        ) = map(partial(self._get_sets, recursive=True), [a, b, c])
        assert up_a_rec == set()
        assert down_a_rec == {b, c}
        assert up_b_rec == {a}
        assert down_b_rec == {c}
        assert up_c_rec == {a, b}
        assert down_c_rec == set()

    @staticmethod
    def _get_sets(
        task: Task, /, *, recursive: bool = False
    ) -> tuple[set[Task], set[Task]]:
        return set(get_dependencies_upstream(task, recursive=recursive)), set(
            get_dependencies_downstream(task, recursive=recursive)
        )


class TestGetTaskClasses:
    @given(namespace_mixin=namespace_mixins())
    @settings(max_examples=1)
    def test_main(self, namespace_mixin: Any) -> None:
        class Example(namespace_mixin, Task):
            ...

        assert Example in get_task_classes()

    def test_notifications(self) -> None:
        assert smtp not in _yield_task_classes()

    @given(namespace_mixin=namespace_mixins())
    @settings(max_examples=1)
    def test_filter(self, namespace_mixin: Any) -> None:
        class Parent(namespace_mixin, Task):
            ...

        class Child(Parent):
            ...

        result = get_task_classes(cls=Parent)
        expected = frozenset([Child])
        assert result == expected


class TestPathTarget:
    def test_main(self, tmp_path: Path) -> None:
        target = PathTarget(path := tmp_path.joinpath("file"))
        assert isinstance(target.path, Path)
        assert not target.exists()
        path.touch()
        assert target.exists()


class TestTimeParameter:
    @given(data=data(), time=times())
    def test_main(self, data: DataObject, time: dt.time) -> None:
        param = TimeParameter()
        input_ = data.draw(sampled_from([time, serialize_time(time)]))
        norm = param.normalize(input_)
        assert param.parse(param.serialize(norm)) == time


class TestWeekdayParameter:
    @given(data=data(), rounding=sampled_from(["prev", "next"]), date=dates())
    def test_main(
        self, data: DataObject, rounding: Literal["prev", "next"], date: dt.date
    ) -> None:
        param = WeekdayParameter(rounding=rounding)
        input_ = data.draw(sampled_from([date, serialize_date(date)]))
        norm = param.normalize(input_)
        assert param.parse(param.serialize(norm)) == norm



from __future__ import annotations

from functools import cache
from typing import Any
from typing import cast

import numpy as np
from numpy import array
from numpy.testing import assert_equal
from pytest import mark
from pytest import param
from pytest import raises
from pytest import skip

from utilities.numpy import NDArrayF

try:
    import cvxpy
except ModuleNotFoundError:
    skip(reason="`cvxpy` missing", allow_module_level=True)


from cvxpy import Expression
from cvxpy import Maximize
from cvxpy import Minimize
from cvxpy import Problem
from cvxpy import Variable

from utilities.cvxpy import InfeasibleProblemError
from utilities.cvxpy import UnboundedProblemError
from utilities.cvxpy import abs_
from utilities.cvxpy import add
from utilities.cvxpy import divide
from utilities.cvxpy import multiply
from utilities.cvxpy import neg
from utilities.cvxpy import norm
from utilities.cvxpy import pos
from utilities.cvxpy import power
from utilities.cvxpy import quad_form
from utilities.cvxpy import solve
from utilities.cvxpy import sqrt
from utilities.cvxpy import subtract
from utilities.cvxpy import sum_


@cache
def _get_variable(
    objective: type[Maximize, Minimize], /, *, array: bool = False
) -> Variable:
    if array:
        var = Variable(2)
        scalar = cvxpy.sum(var)
    else:
        var = Variable()
        scalar = var
    threshold = 10.0
    problem = Problem(
        objective(scalar),
        [cast(Any, var) >= -threshold, cast(Any, var) <= threshold],
    )
    _ = problem.solve()
    return var


class TestAbs:
    @mark.parametrize(
        ("x", "expected"),
        [
            param(0.0, 0.0),
            param(1.0, 1.0),
            param(-1.0, 1.0),
            param(array([0.0]), array([0.0])),
            param(array([1.0]), array([1.0])),
            param(array([-1.0]), array([1.0])),
        ],
    )
    def test_float_and_array(
        self, x: float | NDArrayF, expected: float | NDArrayF
    ) -> None:
        assert_equal(abs_(x), expected)

    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    def test_expression(self, objective: type[Maximize | Minimize]) -> None:
        var = _get_variable(objective)
        assert_equal(abs_(var).value, abs_(var.value))


class TestAdd:
    @mark.parametrize(
        ("x", "y", "expected"),
        [
            param(1.0, 2.0, 3.0),
            param(1.0, array([2.0]), array([3.0])),
            param(array([1.0]), 2.0, array([3.0])),
            param(array([1.0]), array([2.0]), array([3.0])),
        ],
    )
    def test_float_and_array(
        self,
        x: float | NDArrayF,
        y: float | NDArrayF,
        expected: float | NDArrayF,
    ) -> None:
        assert_equal(add(x, y), expected)

    @mark.parametrize("x", [param(1.0), param(array([1.0]))])
    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    def test_one_expression(
        self,
        x: float | NDArrayF | Expression,
        objective: type[Maximize | Minimize],
    ) -> None:
        var = _get_variable(objective)
        assert_equal(add(x, var).value, add(x, var.value))
        assert_equal(add(var, x).value, add(var.value, x))

    @mark.parametrize("objective1", [param(Maximize), param(Minimize)])
    @mark.parametrize("objective2", [param(Maximize), param(Minimize)])
    def test_two_expressions(
        self,
        objective1: type[Maximize | Minimize],
        objective2: type[Maximize | Minimize],
    ) -> None:
        var1 = _get_variable(objective1)
        var2 = _get_variable(objective2)
        assert_equal(add(var1, var2).value, add(var1.value, var2.value))


class TestDivide:
    @mark.parametrize(
        ("x", "y", "expected"),
        [
            param(1.0, 2.0, 0.5),
            param(1.0, array([2.0]), array([0.5])),
            param(array([1.0]), 2.0, array([0.5])),
            param(array([1.0]), array([2.0]), array([0.5])),
        ],
    )
    def test_float_and_array(
        self,
        x: float | NDArrayF,
        y: float | NDArrayF,
        expected: float | NDArrayF,
    ) -> None:
        assert_equal(divide(x, y), expected)

    @mark.parametrize("x", [param(1.0), param(array([1.0]))])
    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    def test_one_expression(
        self,
        x: float | NDArrayF | Expression,
        objective: type[Maximize | Minimize],
    ) -> None:
        var = _get_variable(objective)
        assert_equal(divide(x, var).value, divide(x, var.value))
        assert_equal(divide(var, x).value, divide(var.value, x))

    @mark.parametrize("objective1", [param(Maximize), param(Minimize)])
    @mark.parametrize("objective2", [param(Maximize), param(Minimize)])
    def test_two_expressions(
        self,
        objective1: type[Maximize | Minimize],
        objective2: type[Maximize | Minimize],
    ) -> None:
        var1 = _get_variable(objective1)
        var2 = _get_variable(objective2)
        assert_equal(divide(var1, var2).value, divide(var1.value, var2.value))


class TestMultiply:
    @mark.parametrize(
        ("x", "y", "expected"),
        [
            param(2.0, 3.0, 6.0),
            param(2.0, array([3.0]), array([6.0])),
            param(array([2.0]), 3.0, array([6.0])),
            param(array([2.0]), array([3.0]), array([6.0])),
        ],
    )
    def test_float_and_array(
        self,
        x: float | NDArrayF,
        y: float | NDArrayF,
        expected: float | NDArrayF,
    ) -> None:
        assert_equal(multiply(x, y), expected)

    @mark.parametrize("x", [param(2.0), param(array([2.0]))])
    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    def test_one_expression(
        self,
        x: float | NDArrayF | Expression,
        objective: type[Maximize | Minimize],
    ) -> None:
        var = _get_variable(objective)
        assert_equal(multiply(x, var).value, multiply(x, var.value))
        assert_equal(multiply(var, x).value, multiply(var.value, x))

    @mark.parametrize("objective1", [param(Maximize), param(Minimize)])
    @mark.parametrize("objective2", [param(Maximize), param(Minimize)])
    def test_two_expressions(
        self,
        objective1: type[Maximize | Minimize],
        objective2: type[Maximize | Minimize],
    ) -> None:
        var1 = _get_variable(objective1)
        var2 = _get_variable(objective2)
        assert_equal(
            multiply(var1, var2).value, multiply(var1.value, var2.value)
        )


class TestNeg:
    @mark.parametrize(
        ("x", "expected"),
        [
            param(0.0, 0.0),
            param(1.0, 0.0),
            param(-1.0, 1.0),
            param(array([0.0]), array([0.0])),
            param(array([1.0]), array([0.0])),
            param(array([-1.0]), array([1.0])),
        ],
    )
    def test_float_and_array(
        self, x: float | NDArrayF, expected: float | NDArrayF
    ) -> None:
        assert_equal(neg(x), expected)

    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    def test_expression(self, objective: type[Maximize, Minimize]) -> None:
        var = _get_variable(objective)
        assert_equal(neg(var).value, neg(var.value))


class TestNorm:
    def test_array(self) -> None:
        assert_equal(norm(array([2.0, 3.0])), np.sqrt(13))

    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    def test_expression(self, objective: type[Maximize, Minimize]) -> None:
        var = _get_variable(objective, array=True)
        assert_equal(norm(var).value, norm(var.value))


class TestPos:
    @mark.parametrize(
        ("x", "expected"),
        [
            param(0.0, 0.0),
            param(1.0, 1.0),
            param(-1.0, 0.0),
            param(array([0.0]), array([0.0])),
            param(array([1.0]), array([1.0])),
            param(array([-1.0]), array([0.0])),
        ],
    )
    def test_float_and_array(
        self, x: float | NDArrayF, expected: float | NDArrayF
    ) -> None:
        assert_equal(pos(x), expected)

    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    def test_expression(self, objective: type[Maximize | Minimize]) -> None:
        var = _get_variable(objective)
        assert_equal(pos(var).value, pos(var.value))


class TestPower:
    @mark.parametrize(
        ("x", "p", "expected"),
        [
            param(0.0, 0.0, 1.0),
            param(2.0, 3.0, 8.0),
            param(2.0, array([3.0]), array([8.0])),
            param(array([2.0]), 3.0, array([8.0])),
            param(array([2.0]), array([3.0]), array([8.0])),
        ],
    )
    def test_float_and_array(
        self,
        x: float | NDArrayF,
        p: float | NDArrayF,
        expected: float | NDArrayF,
    ) -> None:
        assert_equal(power(x, p), expected)

    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    def test_one_expression(self, objective: type[Maximize | Minimize]) -> None:
        var = _get_variable(objective)
        assert_equal(power(var, 2.0).value, power(var.value, 2.0))


class TestQuadForm:
    def test_array(self) -> None:
        assert_equal(
            quad_form(array([2.0, 3.0]), array([[4.0, 5.0], [5.0, 4.0]])), 112.0
        )

    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    def test_expression(self, objective: type[Maximize | Minimize]) -> None:
        var = _get_variable(objective, array=True)
        P = array([[2.0, 3.0], [3.0, 2.0]])  # noqa: N806
        assert_equal(quad_form(var, P).value, quad_form(var.value, P))


class TestSolve:
    def test_main(self) -> None:
        var = Variable()
        problem = Problem(Minimize(sum_(abs_(var))), [])
        _ = solve(problem)

    def test_infeasible_problem(self) -> None:
        var = Variable()
        threshold = 1.0
        problem = Problem(
            Minimize(sum_(abs_(var))),
            [cast(Any, var) >= threshold, cast(Any, var) <= -threshold],
        )
        with raises(InfeasibleProblemError):
            _ = solve(problem)

    def test_unbounded_problem(self) -> None:
        var = Variable()
        problem = Problem(Maximize(sum_(var)), [])
        with raises(UnboundedProblemError):
            _ = solve(problem)


class TestSqrt:
    @mark.parametrize(
        ("x", "expected"),
        [
            param(0.0, 0.0),
            param(1.0, 1.0),
            param(array([0.0]), array([0.0])),
            param(array([1.0]), array([1.0])),
        ],
    )
    def test_float_and_array(
        self, x: float | NDArrayF, expected: float | NDArrayF
    ) -> None:
        assert_equal(sqrt(x), expected)

    def test_expression(self) -> None:
        var = _get_variable(Maximize)
        assert_equal(sqrt(var).value, sqrt(var.value))


class TestSubtract:
    @mark.parametrize(
        ("x", "y", "expected"),
        [
            param(1.0, 2.0, -1.0),
            param(1.0, array([2.0]), array([-1.0])),
            param(array([1.0]), 2.0, array([-1.0])),
            param(array([1.0]), array([2.0]), array([-1.0])),
        ],
    )
    def test_float_and_array(
        self,
        x: float | NDArrayF,
        y: float | NDArrayF,
        expected: float | NDArrayF,
    ) -> None:
        assert_equal(subtract(x, y), expected)

    @mark.parametrize("x", [param(1.0), param(array([1.0]))])
    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    def test_one_expression(
        self,
        x: float | NDArrayF | Expression,
        objective: type[Maximize | Minimize],
    ) -> None:
        var = _get_variable(objective)
        assert_equal(subtract(x, var).value, subtract(x, var.value))
        assert_equal(subtract(var, x).value, subtract(var.value, x))

    @mark.parametrize("objective1", [param(Maximize), param(Minimize)])
    @mark.parametrize("objective2", [param(Maximize), param(Minimize)])
    def test_two_expressions(
        self,
        objective1: type[Maximize | Minimize],
        objective2: type[Maximize | Minimize],
    ) -> None:
        var1 = _get_variable(objective1)
        var2 = _get_variable(objective2)
        assert_equal(
            subtract(var1, var2).value, subtract(var1.value, var2.value)
        )


class TestSum:
    @mark.parametrize(
        ("x", "expected"),
        [
            param(0.0, 0.0),
            param(1.0, 1.0),
            param(-1.0, -1.0),
            param(array([0.0]), 0.0),
            param(array([1.0]), 1.0),
            param(array([-1.0]), -1.0),
        ],
    )
    def test_float_and_array(
        self, x: float | NDArrayF, expected: float
    ) -> None:
        assert_equal(sum_(x), expected)

    def test_expression(self) -> None:
        var = _get_variable(Maximize)
        assert_equal(sum_(var).value, sum_(var.value))



from __future__ import annotations

import datetime as dt
from collections.abc import Callable
from typing import TYPE_CHECKING
from typing import Any
from typing import cast

from hypothesis import assume
from hypothesis import given
from hypothesis.extra.pandas import range_indexes
from hypothesis.strategies import integers
from numpy import array
from numpy import nan
from numpy.testing import assert_equal
from pandas import NA
from pandas import DataFrame
from pandas import Index
from pandas import NaT
from pandas import RangeIndex
from pandas import Series
from pandas import Timestamp
from pandas import concat
from pandas import to_datetime
from pandas.testing import assert_series_equal
from pytest import mark
from pytest import param
from pytest import raises

from utilities.datetime import TODAY
from utilities.datetime import UTC
from utilities.hypothesis import text_ascii
from utilities.hypothesis import timestamps
from utilities.numpy import datetime64ns
from utilities.pandas import TIMESTAMP_MAX_AS_DATE
from utilities.pandas import TIMESTAMP_MAX_AS_DATETIME
from utilities.pandas import TIMESTAMP_MIN_AS_DATE
from utilities.pandas import TIMESTAMP_MIN_AS_DATETIME
from utilities.pandas import DataFrameColumnsError
from utilities.pandas import DataFrameColumnsNameError
from utilities.pandas import DataFrameDTypesError
from utilities.pandas import DataFrameRangeIndexError
from utilities.pandas import DifferentDTypeError
from utilities.pandas import EmptyPandasConcatError
from utilities.pandas import Int64
from utilities.pandas import RangeIndexNameError
from utilities.pandas import RangeIndexStartError
from utilities.pandas import RangeIndexStepError
from utilities.pandas import SeriesRangeIndexError
from utilities.pandas import TimestampIsNaTError
from utilities.pandas import boolean
from utilities.pandas import check_dataframe
from utilities.pandas import check_range_index
from utilities.pandas import redirect_to_empty_pandas_concat_error
from utilities.pandas import series_max
from utilities.pandas import series_min
from utilities.pandas import string
from utilities.pandas import timestamp_to_date
from utilities.pandas import timestamp_to_datetime
from utilities.pandas import to_numpy

if TYPE_CHECKING:  # pragma: no cover
    from utilities.pandas import SeriesA


class TestCheckDataFrame:
    def test_main(self) -> None:
        df = DataFrame(index=RangeIndex(0))
        check_dataframe(df)

    def test_columns_name_error(self) -> None:
        df = DataFrame(
            0.0, index=RangeIndex(1), columns=Index(["value"], name="name")
        )
        with raises(DataFrameColumnsNameError):
            check_dataframe(df)

    def test_columns_pass(self) -> None:
        df = DataFrame(0.0, index=RangeIndex(1), columns=["value"])
        check_dataframe(df, columns=["value"])

    def test_columns_error(self) -> None:
        df = DataFrame(0.0, index=RangeIndex(1), columns=["value"])
        with raises(DataFrameColumnsError):
            check_dataframe(df, columns=["other"])

    def test_dtypes_pass(self) -> None:
        df = DataFrame(0.0, index=RangeIndex(1), columns=["value"])
        check_dataframe(df, dtypes={"value": float})

    def test_dtypes_error(self) -> None:
        df = DataFrame(0.0, index=RangeIndex(1), columns=["value"])
        with raises(DataFrameDTypesError):
            check_dataframe(df, dtypes={"value": int})


class TestCheckRangeIndex:
    @given(index=range_indexes())
    def test_main(self, index: RangeIndex) -> None:
        check_range_index(index)

    def test_type(self) -> None:
        index = Index([], dtype=float)
        with raises(TypeError):
            check_range_index(index)

    @given(start=integers(-10, 10), stop=integers(-10, 10))
    def test_start(self, start: int, stop: int) -> None:
        _ = assume(start != 0)
        index = RangeIndex(start=start, stop=stop)
        with raises(RangeIndexStartError):
            check_range_index(index)

    @given(step=integers(-10, 10))
    def test_step(self, step: int) -> None:
        _ = assume(step not in {0, 1})
        index = RangeIndex(step=step)
        with raises(RangeIndexStepError):
            check_range_index(index)

    @given(index=range_indexes(name=text_ascii()))
    def test_name(self, index: RangeIndex) -> None:
        with raises(RangeIndexNameError):
            check_range_index(index)

    def test_series_pass(self) -> None:
        series = Series(index=RangeIndex(0), dtype=float)
        check_range_index(series)

    def test_series_fail(self) -> None:
        series = Series(index=Index([], dtype=int), dtype=float)
        with raises(SeriesRangeIndexError):
            check_range_index(series)

    def test_dataframe_pass(self) -> None:
        df = DataFrame(index=RangeIndex(0))
        check_range_index(df)

    def test_dataframe_fail(self) -> None:
        df = DataFrame(index=Index([], dtype=int))
        with raises(DataFrameRangeIndexError):
            check_range_index(df)


class TestDTypes:
    @mark.parametrize("dtype", [param(Int64), param(boolean), param(string)])
    def test_main(self, dtype: Any) -> None:
        assert isinstance(Series([], dtype=dtype), Series)


class TestRedirectToEmptyPandasConcatError:
    def test_main(self) -> None:
        with raises(EmptyPandasConcatError):
            try:
                _ = concat([])
            except ValueError as error:
                redirect_to_empty_pandas_concat_error(error)


class TestSeriesMinMax:
    @mark.parametrize(
        ("x_v", "y_v", "dtype", "expected_min_v", "expected_max_v"),
        [
            param(0.0, 1.0, float, 0.0, 1.0),
            param(0.0, nan, float, 0.0, 0.0),
            param(nan, 1.0, float, 1.0, 1.0),
            param(nan, nan, float, nan, nan),
            param(0, 1, Int64, 0, 1),
            param(0, NA, Int64, 0, 0),
            param(NA, 1, Int64, 1, 1),
            param(NA, NA, Int64, NA, NA),
            param(
                TIMESTAMP_MIN_AS_DATE,
                TIMESTAMP_MAX_AS_DATE,
                datetime64ns,
                TIMESTAMP_MIN_AS_DATE,
                TIMESTAMP_MAX_AS_DATE,
            ),
            param(
                TIMESTAMP_MIN_AS_DATE,
                NaT,
                datetime64ns,
                TIMESTAMP_MIN_AS_DATE,
                TIMESTAMP_MIN_AS_DATE,
            ),
            param(
                NaT,
                TIMESTAMP_MAX_AS_DATE,
                datetime64ns,
                TIMESTAMP_MAX_AS_DATE,
                TIMESTAMP_MAX_AS_DATE,
            ),
            param(NaT, NaT, datetime64ns, NaT, NaT),
        ],
    )
    def test_main(
        self,
        x_v: Any,
        y_v: Any,
        dtype: Any,
        expected_min_v: Any,
        expected_max_v: Any,
    ) -> None:
        x = Series(data=[x_v], dtype=dtype)
        y = Series(data=[y_v], dtype=dtype)
        result_min = series_min(x, y)
        expected_min = Series(data=[expected_min_v], dtype=dtype)
        assert_series_equal(result_min, expected_min)
        result_max = series_max(x, y)
        expected_max = Series(data=[expected_max_v], dtype=dtype)
        assert_series_equal(result_max, expected_max)

    @mark.parametrize("func", [param(series_min), param(series_max)])
    def test_different_index(
        self, func: Callable[[SeriesA, SeriesA], SeriesA]
    ) -> None:
        x = Series(data=nan, index=Index([0], dtype=int))
        y = Series(data=nan, index=Index([1], dtype=int))
        with raises(AssertionError):
            _ = func(x, y)

    @mark.parametrize("func", [param(series_min), param(series_max)])
    def test_different_dtype(
        self, func: Callable[[SeriesA, SeriesA], SeriesA]
    ) -> None:
        x = Series(data=nan, dtype=float)
        y = Series(data=NA, dtype=Int64)
        with raises(DifferentDTypeError):
            _ = func(x, y)


class TestTimestampMinMaxAsDate:
    def test_min(self) -> None:
        date = TIMESTAMP_MIN_AS_DATE
        assert isinstance(to_datetime(cast(Timestamp, date)), Timestamp)
        with raises(ValueError, match="Out of bounds nanosecond timestamp"):
            _ = to_datetime(cast(Timestamp, date - dt.timedelta(days=1)))

    def test_max(self) -> None:
        date = TIMESTAMP_MAX_AS_DATE
        assert isinstance(to_datetime(cast(Timestamp, date)), Timestamp)
        with raises(ValueError, match="Out of bounds nanosecond timestamp"):
            _ = to_datetime(cast(Timestamp, date + dt.timedelta(days=1)))


class TestTimestampMinMaxAsDateTime:
    def test_min(self) -> None:
        date = TIMESTAMP_MIN_AS_DATETIME
        assert isinstance(to_datetime(date), Timestamp)
        with raises(ValueError, match="Out of bounds nanosecond timestamp"):
            _ = to_datetime(date - dt.timedelta(microseconds=1))

    def test_max(self) -> None:
        date = TIMESTAMP_MAX_AS_DATETIME
        assert isinstance(to_datetime(date), Timestamp)
        with raises(ValueError, match="Out of bounds nanosecond timestamp"):
            _ = to_datetime(date + dt.timedelta(microseconds=1))


class TestTimestampToDate:
    @mark.parametrize(
        ("timestamp", "expected"),
        [
            param(to_datetime("2000-01-01"), dt.date(2000, 1, 1)),
            param(to_datetime("2000-01-01 12:00:00"), dt.date(2000, 1, 1)),
        ],
    )
    def test_main(self, timestamp: Any, expected: dt.date) -> None:
        assert timestamp_to_date(timestamp) == expected

    def test_error(self) -> None:
        with raises(TimestampIsNaTError):
            _ = timestamp_to_date(NaT)


class TestTimestampToDateTime:
    @mark.parametrize(
        ("timestamp", "expected"),
        [
            param(
                to_datetime("2000-01-01"), dt.datetime(2000, 1, 1, tzinfo=UTC)
            ),
            param(
                to_datetime("2000-01-01 12:00:00"),
                dt.datetime(2000, 1, 1, 12, tzinfo=UTC),
            ),
            param(
                to_datetime("2000-01-01 12:00:00+00:00"),
                dt.datetime(2000, 1, 1, 12, tzinfo=UTC),
            ),
        ],
    )
    def test_main(self, timestamp: Any, expected: dt.datetime) -> None:
        assert timestamp_to_datetime(timestamp) == expected

    @given(timestamp=timestamps(allow_nanoseconds=True))
    def test_warn(self, timestamp: Timestamp) -> None:
        _ = assume(cast(Any, timestamp).nanosecond != 0)
        with raises(
            UserWarning, match="Discarding nonzero nanoseconds in conversion"
        ):
            _ = timestamp_to_datetime(timestamp)

    def test_error(self) -> None:
        with raises(TimestampIsNaTError):
            _ = timestamp_to_datetime(NaT)


class TestToNumpy:
    @mark.parametrize(
        ("series_v", "series_d", "array_v", "array_d"),
        [
            param(True, bool, True, bool),
            param(False, bool, False, bool),
            param(True, boolean, True, object),
            param(False, boolean, False, object),
            param(NA, boolean, None, object),
            param(TODAY, datetime64ns, TODAY, datetime64ns),
            param(0, int, 0, int),
            param(0, Int64, 0, object),
            param(NA, Int64, None, object),
            param(nan, float, nan, float),
            param("", string, "", object),
            param(NA, string, None, object),
        ],
    )
    def test_main(
        self, series_v: Any, series_d: Any, array_v: Any, array_d: Any
    ) -> None:
        series = Series([series_v], dtype=series_d)
        result = to_numpy(series)
        expected = array([array_v], dtype=array_d)
        assert_equal(result, expected)



from __future__ import annotations

from hypothesis import given
from hypothesis.strategies import DataObject
from hypothesis.strategies import data
from hypothesis.strategies import floats
from hypothesis.strategies import integers

from utilities.hypothesis import assume_does_not_raise
from utilities.hypothesis import float_arrays
from utilities.numpy import NDArrayF
from utilities.numpy import ewma
from utilities.numpy import exp_moving_sum


class TestEwma:
    @given(data=data(), array=float_arrays(), halflife=floats(0.1, 10.0))
    def test_main(
        self, data: DataObject, array: NDArrayF, halflife: float
    ) -> None:
        axis = data.draw(integers(0, array.ndim - 1)) if array.ndim >= 1 else -1
        with assume_does_not_raise(RuntimeWarning):
            _ = ewma(array, halflife, axis=axis)


class TestExpMovingSum:
    @given(data=data(), array=float_arrays(), halflife=floats(0.1, 10.0))
    def test_main(
        self, data: DataObject, array: NDArrayF, halflife: float
    ) -> None:
        axis = data.draw(integers(0, array.ndim - 1)) if array.ndim >= 1 else -1
        with assume_does_not_raise(RuntimeWarning):
            _ = exp_moving_sum(array, halflife, axis=axis)



from __future__ import annotations

from enum import Enum
from enum import auto
from typing import cast

from hypothesis import given
from hypothesis.strategies import DataObject
from hypothesis.strategies import data
from hypothesis.strategies import sampled_from
from pytest import raises

from utilities.enum import MultipleMatchingMembersError
from utilities.enum import NoMatchingMemberError
from utilities.enum import StrEnum
from utilities.enum import ensure_enum
from utilities.enum import parse_enum


class TestParseEnum:
    @given(data=data())
    def test_main(self, data: DataObject) -> None:
        class Truth(Enum):
            true = auto()
            false = auto()

        truth = data.draw(sampled_from(Truth))
        result = parse_enum(Truth, truth.name)
        assert result is truth

    @given(data=data())
    def test_case_insensitive(self, data: DataObject) -> None:
        class Truth(Enum):
            true = auto()
            false = auto()

        truth = data.draw(sampled_from(Truth))
        name = truth.name
        input_ = data.draw(sampled_from([name, name.upper(), name.lower()]))
        result = parse_enum(Truth, input_, case_sensitive=False)
        assert result is truth

    def test_no_matching_member(self) -> None:
        class Example(Enum):
            pass

        with raises(NoMatchingMemberError):
            _ = parse_enum(Example, "not-a-member")

    @given(data=data())
    def test_multiple_matching_members(self, data: DataObject) -> None:
        class Example(Enum):
            member = auto()
            MEMBER = auto()

        member = data.draw(sampled_from(Example))
        with raises(MultipleMatchingMembersError):
            _ = parse_enum(Example, member.name, case_sensitive=False)


class TestEnsureEnum:
    @given(data=data())
    def test_main(self, data: DataObject) -> None:
        class Truth(Enum):
            true = auto()
            false = auto()

        truth = data.draw(sampled_from(Truth))
        input_ = data.draw(sampled_from([truth, truth.name]))
        result = ensure_enum(Truth, input_)
        assert result is truth


class TestStrEnum:
    @given(data=data())
    def test_main(self, data: DataObject) -> None:
        class Truth(cast(type[Enum], StrEnum)):
            true = auto()
            false = auto()

        truth = data.draw(sampled_from(Truth))
        assert isinstance(truth, Enum)
        assert isinstance(truth, str)
        assert truth == truth.name



from __future__ import annotations

from pathlib import Path
from typing import Any

from hypothesis import given
from hypothesis.strategies import DataObject
from hypothesis.strategies import data
from hypothesis.strategies import integers
from hypothesis.strategies import sets
from sqlalchemy import Column
from sqlalchemy import Engine
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import Table
from sqlalchemy import insert
from sqlalchemy import select
from sqlalchemy.orm import declarative_base

from utilities.hypothesis import sqlite_engines


class TestSQLiteEngines:
    @given(engine=sqlite_engines())
    def test_main(self, engine: Engine) -> None:
        assert isinstance(engine, Engine)
        assert (database := engine.url.database) is not None
        assert not Path(database).exists()

    @given(data=data(), ids=sets(integers(0, 100), min_size=1, max_size=10))
    def test_core(self, data: DataObject, ids: set[int]) -> None:
        metadata = MetaData()
        table = Table(
            "example", metadata, Column("id_", Integer, primary_key=True)
        )
        engine = data.draw(sqlite_engines(metadata=metadata))
        self._run_test(engine, table, ids)

    @given(data=data(), ids=sets(integers(0, 100), min_size=1, max_size=10))
    def test_orm(self, data: DataObject, ids: set[int]) -> None:
        Base = declarative_base()  # noqa: N806

        class Example(Base):
            __tablename__ = "example"

            id_ = Column(Integer, primary_key=True)

        engine = data.draw(sqlite_engines(base=Base))
        self._run_test(engine, Example, ids)

    def _run_test(
        self, engine: Engine, table_or_model: Any, ids: set[int]
    ) -> None:
        with engine.begin() as conn:
            _ = conn.execute(
                insert(table_or_model), [{"id_": id_} for id_ in ids]
            )
            res = conn.execute(select(table_or_model)).all()
        assert {r.id_ for r in res} == ids



import datetime as dt
import enum
from collections.abc import Callable
from enum import auto
from pathlib import Path
from subprocess import check_call
from typing import Any
from typing import cast

from click import command
from click import echo
from click.testing import CliRunner
from hypothesis import given
from hypothesis.strategies import DataObject
from hypothesis.strategies import SearchStrategy
from hypothesis.strategies import data
from hypothesis.strategies import dates
from hypothesis.strategies import datetimes
from hypothesis.strategies import just
from hypothesis.strategies import sampled_from
from hypothesis.strategies import timedeltas
from hypothesis.strategies import times
from hypothesis.strategies import tuples
from pytest import MonkeyPatch
from pytest import mark
from pytest import param
from pytest import raises
from sqlalchemy import Engine
from typed_settings import settings
from typed_settings.exceptions import InvalidValueError

from utilities.datetime import UTC
from utilities.datetime import serialize_date
from utilities.datetime import serialize_datetime
from utilities.datetime import serialize_time
from utilities.datetime import serialize_timedelta
from utilities.hypothesis import sqlite_engines
from utilities.hypothesis import temp_paths
from utilities.hypothesis import text_ascii
from utilities.sqlalchemy import serialize_engine
from utilities.typed_settings import AppNameContainsUnderscoreError
from utilities.typed_settings import _get_loaders
from utilities.typed_settings import click_options
from utilities.typed_settings import get_repo_root_config
from utilities.typed_settings import load_settings

app_names = text_ascii(min_size=1).map(str.lower)


class TestGetRepoRootConfig:
    def test_exists(self, monkeypatch: MonkeyPatch, tmp_path: Path) -> None:
        monkeypatch.chdir(tmp_path)
        _ = check_call(["git", "init"])  # noqa: S603, S607
        Path("config.toml").touch()
        expected = tmp_path.joinpath("config.toml")
        assert get_repo_root_config(cwd=tmp_path) == expected

    def test_does_not_exist(self, tmp_path: Path) -> None:
        assert get_repo_root_config(cwd=tmp_path) is None


class TestGetLoaders:
    def test_success(self) -> None:
        _ = _get_loaders()

    def test_error(self) -> None:
        with raises(AppNameContainsUnderscoreError):
            _ = _get_loaders(appname="app_name")


class TestLoadSettings:
    @given(data=data(), appname=app_names, root=temp_paths())
    @mark.parametrize(
        ("cls", "strategy", "serialize"),
        [
            param(dt.date, dates(), serialize_date),
            param(
                dt.datetime, datetimes(timezones=just(UTC)), serialize_datetime
            ),
            param(dt.time, times(), serialize_time),
            param(dt.timedelta, timedeltas(), serialize_timedelta),
            param(Engine, sqlite_engines(), serialize_engine),
        ],
    )
    def test_main(
        self,
        data: DataObject,
        appname: str,
        root: Path,
        cls: Any,
        strategy: SearchStrategy[Any],
        serialize: Callable[[Any], str],
    ) -> None:
        default, value = data.draw(tuples(strategy, strategy))

        @settings(frozen=True)
        class Settings:
            value: cls = default

        settings_default = load_settings(Settings)
        assert settings_default.value == default
        file = root.joinpath("file.toml")
        with file.open(mode="w") as fh:
            _ = fh.write(f'[{appname}]\nvalue = "{serialize(value)}"')
        settings_loaded = load_settings(
            Settings, appname=appname, config_files=[file]
        )
        try:
            assert settings_loaded.value == value
        except AssertionError:
            assert settings_loaded.value.url == value.url

    @given(appname=app_names)
    @mark.parametrize(
        "cls", [param(dt.date), param(dt.time), param(dt.timedelta)]
    )
    def test_errors(self, appname: str, cls: Any) -> None:
        @settings(frozen=True)
        class Settings:
            value: cls = cast(Any, None)

        with raises(InvalidValueError):
            _ = load_settings(Settings, appname=appname)


class TestClickOptions:
    @given(data=data(), appname=app_names, root=temp_paths())
    @mark.parametrize(
        ("cls", "strategy", "serialize"),
        [
            param(dt.date, dates(), serialize_date),
            param(
                dt.datetime, datetimes(timezones=just(UTC)), serialize_datetime
            ),
            param(dt.time, times(), serialize_time),
            param(dt.timedelta, timedeltas(), serialize_timedelta),
            param(Engine, sqlite_engines(), serialize_engine),
        ],
    )
    def test_main(
        self,
        data: DataObject,
        appname: str,
        root: Path,
        cls: Any,
        strategy: SearchStrategy[Any],
        serialize: Callable[[Any], str],
    ) -> None:
        default, val, cfg = data.draw(tuples(strategy, strategy, strategy))
        val_str, cfg_str = map(serialize, [val, cfg])
        runner = CliRunner()

        @settings(frozen=True)
        class Config:
            value: cls = default

        @command()
        @click_options(Config, appname=appname)
        def cli1(config: Config, /) -> None:
            echo(f"value = {serialize(config.value)}")

        result = runner.invoke(cli1)
        assert result.exit_code == 0
        assert result.stdout == f"value = {serialize(default)}\n"

        result = runner.invoke(cli1, f'--value="{val_str}"')
        assert result.exit_code == 0
        assert result.stdout == f"value = {val_str}\n"

        file = root.joinpath("file.toml")
        with file.open(mode="w") as fh:
            _ = fh.write(f'[{appname}]\nvalue = "{cfg_str}"')

        @command()
        @click_options(Config, appname=appname, config_files=[file])
        def cli2(config: Config, /) -> None:
            echo(f"value = {serialize(config.value)}")

        result = runner.invoke(cli2)
        assert result.exit_code == 0
        assert result.stdout == f"value = {cfg_str}\n"

        result = runner.invoke(cli1, f'--value="{val_str}"')
        assert result.exit_code == 0
        assert result.stdout == f"value = {val_str}\n"

    @given(data=data())
    def test_enum(self, data: DataObject) -> None:
        class Truth(enum.Enum):
            true = auto()
            false = auto()

        default = data.draw(sampled_from(Truth))

        @settings(frozen=True)
        class Config:
            value: Truth = default

        @command()
        @click_options(Config)
        def cli(config: Config, /) -> None:
            echo(f"truth = {config.value.name}")

        result = CliRunner().invoke(cli)
        assert result.exit_code == 0
        assert result.stdout == f"truth = {default.name}\n"

        arg = data.draw(sampled_from(Truth))
        result = CliRunner().invoke(cli, ["--value", arg.name])
        assert result.exit_code == 0
        assert result.stdout == f"truth = {arg.name}\n"



from __future__ import annotations

from threading import get_native_id
from time import sleep
from typing import Any
from typing import cast

from click import command
from click import option
from loguru import logger
from luigi import IntParameter
from luigi import Task
from numpy import arange
from numpy import array
from numpy.random import default_rng
from numpy.random import random
from typing_extensions import override

from utilities.atomicwrites import writer
from utilities.class_name import get_class_name
from utilities.logging import LogLevel
from utilities.loguru import setup_loguru
from utilities.luigi import PathTarget
from utilities.luigi import build
from utilities.os import CPU_COUNT
from utilities.tempfile import TEMP_DIR


class Example(Task):
    """Example task."""

    messages = cast(int, IntParameter())

    @override
    def output(self) -> PathTarget:  # type: ignore
        return PathTarget(TEMP_DIR.joinpath(get_class_name(self)))

    @override
    def run(self) -> None:
        rng = default_rng(get_native_id())
        levels = [level.name for level in LogLevel]
        p = array([40, 30, 20, 9, 1]) / 100.0
        for i in arange(n := self.messages) + 1:
            level = rng.choice(levels, p=p)
            logger.log(level, "{}, #{}/{}", get_class_name(self), i, n)
            sleep(1.0 + random())
        with writer(self.output().path) as temp:
            temp.touch()


@command()
@option("-t", "--tasks", default=10)
@option("-m", "--messages", default=60)
def main(*, tasks: int, messages: int) -> None:
    """Run the test script."""
    setup_loguru(levels={"luigi": LogLevel.DEBUG}, files="test_luigi")
    classes = [type(f"Example{i}", (Example,), {}) for i in range(tasks)]
    instances = [
        cast(Example, cast(Any, cls)(messages=messages)) for cls in classes
    ]
    _ = build(instances, local_scheduler=True, workers=CPU_COUNT)


if __name__ == "__main__":
    main()



from __future__ import annotations

from collections.abc import Iterator
from contextlib import contextmanager
from os import chdir
from pathlib import Path

from utilities.re import extract_group

PathLike = Path | str


def ensure_suffix(path: PathLike, suffix: str, /) -> Path:
    """Ensure a path has the required suffix."""
    as_path = Path(path)
    parts = as_path.name.split(".")
    clean_suffix = extract_group(r"^\.(\w+)$", suffix)
    if parts[-1] != clean_suffix:
        parts.append(clean_suffix)
    return as_path.with_name(".".join(parts))


@contextmanager
def temp_cwd(path: PathLike, /) -> Iterator[None]:
    """Context manager with temporary current working directory set."""
    prev = Path.cwd()
    chdir(path)
    try:
        yield
    finally:
        chdir(prev)



from __future__ import annotations

from collections.abc import Callable
from collections.abc import Mapping
from functools import partial
from operator import neg
from operator import pow
from typing import Any
from typing import Literal

from pytest import mark
from pytest import param

from utilities.class_name import get_class_name
from utilities.pqdm import _get_desc
from utilities.pqdm import pmap
from utilities.pqdm import pstarmap
from utilities.sentinel import Sentinel
from utilities.sentinel import sentinel


class TestGetDesc:
    @mark.parametrize(
        ("desc", "func", "expected"),
        [
            param(sentinel, neg, {"desc": "neg"}),
            param(sentinel, partial(neg), {"desc": "neg"}),
            param(None, neg, {}),
            param("custom", neg, {"desc": "custom"}),
        ],
    )
    def test_main(
        self,
        desc: str | None | Sentinel,
        func: Callable[..., Any],
        expected: Mapping[str, str],
    ) -> None:
        assert _get_desc(desc, func) == expected

    def test_class(self) -> None:
        class Example:
            def __call__(self) -> None:
                return

        assert _get_desc(sentinel, Example()) == {
            "desc": get_class_name(Example)
        }


class TestPMap:
    @mark.parametrize("parallelism", [param("processes"), param("threads")])
    @mark.parametrize("n_jobs", [param(1), param(2)])
    def test_unary(
        self, parallelism: Literal["processes", "threads"], n_jobs: int
    ) -> None:
        result = pmap(neg, [1, 2, 3], parallelism=parallelism, n_jobs=n_jobs)
        expected = [-1, -2, -3]
        assert result == expected

    @mark.parametrize("parallelism", [param("processes"), param("threads")])
    @mark.parametrize("n_jobs", [param(1), param(2)])
    def test_binary(
        self, parallelism: Literal["processes", "threads"], n_jobs: int
    ) -> None:
        result = pmap(
            pow, [2, 3, 10], [5, 2, 3], parallelism=parallelism, n_jobs=n_jobs
        )
        expected = [32, 9, 1000]
        assert result == expected


class TestPStarMap:
    @mark.parametrize("parallelism", [param("processes"), param("threads")])
    @mark.parametrize("n_jobs", [param(1), param(2)])
    def test_unary(
        self, parallelism: Literal["processes", "threads"], n_jobs: int
    ) -> None:
        result = pstarmap(
            neg, [(1,), (2,), (3,)], parallelism=parallelism, n_jobs=n_jobs
        )
        expected = [-1, -2, -3]
        assert result == expected

    @mark.parametrize("parallelism", [param("processes"), param("threads")])
    @mark.parametrize("n_jobs", [param(1), param(2)])
    def test_binary(
        self, parallelism: Literal["processes", "threads"], n_jobs: int
    ) -> None:
        result = pstarmap(
            pow,
            [(2, 5), (3, 2), (10, 3)],
            parallelism=parallelism,
            n_jobs=n_jobs,
        )
        expected = [32, 9, 1000]
        assert result == expected



from __future__ import annotations

import datetime as dt
from pathlib import Path
from typing import Any

from pytest import mark
from pytest import param
from pytest import raises

from utilities.datetime import UTC
from utilities.json import serialize


class TestSerialize:
    @mark.parametrize(
        ("x", "expected"),
        [
            param(dt.date(2000, 1, 1), '"2000-01-01"'),
            param(
                dt.datetime(2000, 1, 1, 12, tzinfo=UTC),
                '"2000-01-01T12:00:00+00:00"',
            ),
            param(Path("a/b/c"), '"a/b/c"'),
            param({1, 2, 3}, '"set([1, 2, 3])"'),
            param({"a", "b", "c"}, '"set([\\"a\\", \\"b\\", \\"c\\"])"'),
            param(frozenset([1, 2, 3]), '"frozenset([1, 2, 3])"'),
            param(
                frozenset(["a", "b", "c"]),
                '"frozenset([\\"a\\", \\"b\\", \\"c\\"])"',
            ),
        ],
    )
    def test_main(self, x: Any, expected: str) -> None:
        assert serialize(x) == expected

    def test_error(self) -> None:
        class Example:
            pass

        with raises(TypeError, match="Invalid type"):
            _ = serialize(Example())



from __future__ import annotations

from contextlib import suppress
from typing import Annotated
from typing import Any

from beartype.door import die_if_unbearable
from beartype.roar import BeartypeDoorHintViolation
from hypothesis import Phase
from hypothesis import example
from hypothesis import given
from hypothesis import settings
from numpy import array
from numpy import empty
from numpy import nan
from numpy import zeros
from numpy.typing import NDArray
from pytest import mark
from pytest import param

from utilities.hypothesis import float_arrays
from utilities.hypothesis import int_arrays
from utilities.numpy import NDArray0
from utilities.numpy import NDArray1
from utilities.numpy import NDArray2
from utilities.numpy import NDArray3
from utilities.numpy import NDArrayB
from utilities.numpy import NDArrayB0
from utilities.numpy import NDArrayB1
from utilities.numpy import NDArrayB2
from utilities.numpy import NDArrayB3
from utilities.numpy import NDArrayD0
from utilities.numpy import NDArrayD1
from utilities.numpy import NDArrayD2
from utilities.numpy import NDArrayD3
from utilities.numpy import NDArrayDas
from utilities.numpy import NDArrayDas0
from utilities.numpy import NDArrayDas1
from utilities.numpy import NDArrayDas2
from utilities.numpy import NDArrayDas3
from utilities.numpy import NDArrayDD
from utilities.numpy import NDArrayDD0
from utilities.numpy import NDArrayDD1
from utilities.numpy import NDArrayDD2
from utilities.numpy import NDArrayDD3
from utilities.numpy import NDArrayDfs
from utilities.numpy import NDArrayDfs0
from utilities.numpy import NDArrayDfs1
from utilities.numpy import NDArrayDfs2
from utilities.numpy import NDArrayDfs3
from utilities.numpy import NDArrayDh
from utilities.numpy import NDArrayDh0
from utilities.numpy import NDArrayDh1
from utilities.numpy import NDArrayDh2
from utilities.numpy import NDArrayDh3
from utilities.numpy import NDArrayDM
from utilities.numpy import NDArrayDm
from utilities.numpy import NDArrayDM0
from utilities.numpy import NDArrayDm0
from utilities.numpy import NDArrayDM1
from utilities.numpy import NDArrayDm1
from utilities.numpy import NDArrayDM2
from utilities.numpy import NDArrayDm2
from utilities.numpy import NDArrayDM3
from utilities.numpy import NDArrayDm3
from utilities.numpy import NDArrayDms
from utilities.numpy import NDArrayDms0
from utilities.numpy import NDArrayDms1
from utilities.numpy import NDArrayDms2
from utilities.numpy import NDArrayDms3
from utilities.numpy import NDArrayDns
from utilities.numpy import NDArrayDns0
from utilities.numpy import NDArrayDns1
from utilities.numpy import NDArrayDns2
from utilities.numpy import NDArrayDns3
from utilities.numpy import NDArrayDps
from utilities.numpy import NDArrayDps0
from utilities.numpy import NDArrayDps1
from utilities.numpy import NDArrayDps2
from utilities.numpy import NDArrayDps3
from utilities.numpy import NDArrayDs
from utilities.numpy import NDArrayDs0
from utilities.numpy import NDArrayDs1
from utilities.numpy import NDArrayDs2
from utilities.numpy import NDArrayDs3
from utilities.numpy import NDArrayDus
from utilities.numpy import NDArrayDus0
from utilities.numpy import NDArrayDus1
from utilities.numpy import NDArrayDus2
from utilities.numpy import NDArrayDus3
from utilities.numpy import NDArrayDW
from utilities.numpy import NDArrayDW0
from utilities.numpy import NDArrayDW1
from utilities.numpy import NDArrayDW2
from utilities.numpy import NDArrayDW3
from utilities.numpy import NDArrayDY
from utilities.numpy import NDArrayDY0
from utilities.numpy import NDArrayDY1
from utilities.numpy import NDArrayDY2
from utilities.numpy import NDArrayDY3
from utilities.numpy import NDArrayF
from utilities.numpy import NDArrayF0
from utilities.numpy import NDArrayF0Fin
from utilities.numpy import NDArrayF0FinInt
from utilities.numpy import NDArrayF0FinIntNan
from utilities.numpy import NDArrayF0FinNan
from utilities.numpy import NDArrayF0FinNeg
from utilities.numpy import NDArrayF0FinNegNan
from utilities.numpy import NDArrayF0FinNonNeg
from utilities.numpy import NDArrayF0FinNonNegNan
from utilities.numpy import NDArrayF0FinNonPos
from utilities.numpy import NDArrayF0FinNonPosNan
from utilities.numpy import NDArrayF0FinNonZr
from utilities.numpy import NDArrayF0FinNonZrNan
from utilities.numpy import NDArrayF0FinPos
from utilities.numpy import NDArrayF0FinPosNan
from utilities.numpy import NDArrayF0Int
from utilities.numpy import NDArrayF0IntNan
from utilities.numpy import NDArrayF0Neg
from utilities.numpy import NDArrayF0NegNan
from utilities.numpy import NDArrayF0NonNeg
from utilities.numpy import NDArrayF0NonNegNan
from utilities.numpy import NDArrayF0NonPos
from utilities.numpy import NDArrayF0NonPosNan
from utilities.numpy import NDArrayF0NonZr
from utilities.numpy import NDArrayF0NonZrNan
from utilities.numpy import NDArrayF0Pos
from utilities.numpy import NDArrayF0PosNan
from utilities.numpy import NDArrayF0Zr
from utilities.numpy import NDArrayF0ZrFinNonMic
from utilities.numpy import NDArrayF0ZrFinNonMicNan
from utilities.numpy import NDArrayF0ZrNan
from utilities.numpy import NDArrayF0ZrNonMic
from utilities.numpy import NDArrayF0ZrNonMicNan
from utilities.numpy import NDArrayF1
from utilities.numpy import NDArrayF1Fin
from utilities.numpy import NDArrayF1FinInt
from utilities.numpy import NDArrayF1FinIntNan
from utilities.numpy import NDArrayF1FinNan
from utilities.numpy import NDArrayF1FinNeg
from utilities.numpy import NDArrayF1FinNegNan
from utilities.numpy import NDArrayF1FinNonNeg
from utilities.numpy import NDArrayF1FinNonNegNan
from utilities.numpy import NDArrayF1FinNonPos
from utilities.numpy import NDArrayF1FinNonPosNan
from utilities.numpy import NDArrayF1FinNonZr
from utilities.numpy import NDArrayF1FinNonZrNan
from utilities.numpy import NDArrayF1FinPos
from utilities.numpy import NDArrayF1FinPosNan
from utilities.numpy import NDArrayF1Int
from utilities.numpy import NDArrayF1IntNan
from utilities.numpy import NDArrayF1Neg
from utilities.numpy import NDArrayF1NegNan
from utilities.numpy import NDArrayF1NonNeg
from utilities.numpy import NDArrayF1NonNegNan
from utilities.numpy import NDArrayF1NonPos
from utilities.numpy import NDArrayF1NonPosNan
from utilities.numpy import NDArrayF1NonZr
from utilities.numpy import NDArrayF1NonZrNan
from utilities.numpy import NDArrayF1Pos
from utilities.numpy import NDArrayF1PosNan
from utilities.numpy import NDArrayF1Zr
from utilities.numpy import NDArrayF1ZrFinNonMic
from utilities.numpy import NDArrayF1ZrFinNonMicNan
from utilities.numpy import NDArrayF1ZrNan
from utilities.numpy import NDArrayF1ZrNonMic
from utilities.numpy import NDArrayF1ZrNonMicNan
from utilities.numpy import NDArrayF2
from utilities.numpy import NDArrayF2Fin
from utilities.numpy import NDArrayF2FinInt
from utilities.numpy import NDArrayF2FinIntNan
from utilities.numpy import NDArrayF2FinNan
from utilities.numpy import NDArrayF2FinNeg
from utilities.numpy import NDArrayF2FinNegNan
from utilities.numpy import NDArrayF2FinNonNeg
from utilities.numpy import NDArrayF2FinNonNegNan
from utilities.numpy import NDArrayF2FinNonPos
from utilities.numpy import NDArrayF2FinNonPosNan
from utilities.numpy import NDArrayF2FinNonZr
from utilities.numpy import NDArrayF2FinNonZrNan
from utilities.numpy import NDArrayF2FinPos
from utilities.numpy import NDArrayF2FinPosNan
from utilities.numpy import NDArrayF2Int
from utilities.numpy import NDArrayF2IntNan
from utilities.numpy import NDArrayF2Neg
from utilities.numpy import NDArrayF2NegNan
from utilities.numpy import NDArrayF2NonNeg
from utilities.numpy import NDArrayF2NonNegNan
from utilities.numpy import NDArrayF2NonPos
from utilities.numpy import NDArrayF2NonPosNan
from utilities.numpy import NDArrayF2NonZr
from utilities.numpy import NDArrayF2NonZrNan
from utilities.numpy import NDArrayF2Pos
from utilities.numpy import NDArrayF2PosNan
from utilities.numpy import NDArrayF2Zr
from utilities.numpy import NDArrayF2ZrFinNonMic
from utilities.numpy import NDArrayF2ZrFinNonMicNan
from utilities.numpy import NDArrayF2ZrNan
from utilities.numpy import NDArrayF2ZrNonMic
from utilities.numpy import NDArrayF2ZrNonMicNan
from utilities.numpy import NDArrayF3
from utilities.numpy import NDArrayF3Fin
from utilities.numpy import NDArrayF3FinInt
from utilities.numpy import NDArrayF3FinIntNan
from utilities.numpy import NDArrayF3FinNan
from utilities.numpy import NDArrayF3FinNeg
from utilities.numpy import NDArrayF3FinNegNan
from utilities.numpy import NDArrayF3FinNonNeg
from utilities.numpy import NDArrayF3FinNonNegNan
from utilities.numpy import NDArrayF3FinNonPos
from utilities.numpy import NDArrayF3FinNonPosNan
from utilities.numpy import NDArrayF3FinNonZr
from utilities.numpy import NDArrayF3FinNonZrNan
from utilities.numpy import NDArrayF3FinPos
from utilities.numpy import NDArrayF3FinPosNan
from utilities.numpy import NDArrayF3Int
from utilities.numpy import NDArrayF3IntNan
from utilities.numpy import NDArrayF3Neg
from utilities.numpy import NDArrayF3NegNan
from utilities.numpy import NDArrayF3NonNeg
from utilities.numpy import NDArrayF3NonNegNan
from utilities.numpy import NDArrayF3NonPos
from utilities.numpy import NDArrayF3NonPosNan
from utilities.numpy import NDArrayF3NonZr
from utilities.numpy import NDArrayF3NonZrNan
from utilities.numpy import NDArrayF3Pos
from utilities.numpy import NDArrayF3PosNan
from utilities.numpy import NDArrayF3Zr
from utilities.numpy import NDArrayF3ZrFinNonMic
from utilities.numpy import NDArrayF3ZrFinNonMicNan
from utilities.numpy import NDArrayF3ZrNan
from utilities.numpy import NDArrayF3ZrNonMic
from utilities.numpy import NDArrayF3ZrNonMicNan
from utilities.numpy import NDArrayFFin
from utilities.numpy import NDArrayFFinInt
from utilities.numpy import NDArrayFFinIntNan
from utilities.numpy import NDArrayFFinNan
from utilities.numpy import NDArrayFFinNeg
from utilities.numpy import NDArrayFFinNegNan
from utilities.numpy import NDArrayFFinNonNeg
from utilities.numpy import NDArrayFFinNonNegNan
from utilities.numpy import NDArrayFFinNonPos
from utilities.numpy import NDArrayFFinNonPosNan
from utilities.numpy import NDArrayFFinNonZr
from utilities.numpy import NDArrayFFinNonZrNan
from utilities.numpy import NDArrayFFinPos
from utilities.numpy import NDArrayFFinPosNan
from utilities.numpy import NDArrayFInt
from utilities.numpy import NDArrayFIntNan
from utilities.numpy import NDArrayFNeg
from utilities.numpy import NDArrayFNegNan
from utilities.numpy import NDArrayFNonNeg
from utilities.numpy import NDArrayFNonNegNan
from utilities.numpy import NDArrayFNonPos
from utilities.numpy import NDArrayFNonPosNan
from utilities.numpy import NDArrayFNonZr
from utilities.numpy import NDArrayFNonZrNan
from utilities.numpy import NDArrayFPos
from utilities.numpy import NDArrayFPosNan
from utilities.numpy import NDArrayFZr
from utilities.numpy import NDArrayFZrFinNonMic
from utilities.numpy import NDArrayFZrFinNonMicNan
from utilities.numpy import NDArrayFZrNan
from utilities.numpy import NDArrayFZrNonMic
from utilities.numpy import NDArrayFZrNonMicNan
from utilities.numpy import NDArrayI
from utilities.numpy import NDArrayI0
from utilities.numpy import NDArrayI0Neg
from utilities.numpy import NDArrayI0NonNeg
from utilities.numpy import NDArrayI0NonPos
from utilities.numpy import NDArrayI0NonZr
from utilities.numpy import NDArrayI0Pos
from utilities.numpy import NDArrayI0Zr
from utilities.numpy import NDArrayI1
from utilities.numpy import NDArrayI1Neg
from utilities.numpy import NDArrayI1NonNeg
from utilities.numpy import NDArrayI1NonPos
from utilities.numpy import NDArrayI1NonZr
from utilities.numpy import NDArrayI1Pos
from utilities.numpy import NDArrayI1Zr
from utilities.numpy import NDArrayI2
from utilities.numpy import NDArrayI2Neg
from utilities.numpy import NDArrayI2NonNeg
from utilities.numpy import NDArrayI2NonPos
from utilities.numpy import NDArrayI2NonZr
from utilities.numpy import NDArrayI2Pos
from utilities.numpy import NDArrayI2Zr
from utilities.numpy import NDArrayI3
from utilities.numpy import NDArrayI3Neg
from utilities.numpy import NDArrayI3NonNeg
from utilities.numpy import NDArrayI3NonPos
from utilities.numpy import NDArrayI3NonZr
from utilities.numpy import NDArrayI3Pos
from utilities.numpy import NDArrayI3Zr
from utilities.numpy import NDArrayINeg
from utilities.numpy import NDArrayINonNeg
from utilities.numpy import NDArrayINonPos
from utilities.numpy import NDArrayINonZr
from utilities.numpy import NDArrayIPos
from utilities.numpy import NDArrayIZr
from utilities.numpy import NDArrayO
from utilities.numpy import NDArrayO0
from utilities.numpy import NDArrayO1
from utilities.numpy import NDArrayO2
from utilities.numpy import NDArrayO3
from utilities.numpy import NDim0
from utilities.numpy import NDim1
from utilities.numpy import NDim2
from utilities.numpy import NDim3
from utilities.numpy import datetime64as
from utilities.numpy import datetime64D
from utilities.numpy import datetime64fs
from utilities.numpy import datetime64h
from utilities.numpy import datetime64M
from utilities.numpy import datetime64m
from utilities.numpy import datetime64ms
from utilities.numpy import datetime64ns
from utilities.numpy import datetime64ps
from utilities.numpy import datetime64s
from utilities.numpy import datetime64us
from utilities.numpy import datetime64W
from utilities.numpy import datetime64Y


class TestNDims:
    @mark.parametrize(
        ("ndim", "hint"),
        [param(0, NDim0), param(1, NDim1), param(2, NDim2), param(3, NDim3)],
    )
    def test_main(self, ndim: int, hint: Any) -> None:
        arr = empty(zeros(ndim, dtype=int), dtype=float)
        die_if_unbearable(arr, Annotated[NDArray[Any], hint])


class TestHints:
    @mark.parametrize(
        ("dtype", "hint"),
        [
            param(bool, NDArrayB),
            param(datetime64Y, NDArrayDY),
            param(datetime64M, NDArrayDM),
            param(datetime64W, NDArrayDW),
            param(datetime64D, NDArrayDD),
            param(datetime64h, NDArrayDh),
            param(datetime64m, NDArrayDm),
            param(datetime64s, NDArrayDs),
            param(datetime64ms, NDArrayDms),
            param(datetime64us, NDArrayDus),
            param(datetime64ns, NDArrayDns),
            param(datetime64ps, NDArrayDps),
            param(datetime64fs, NDArrayDfs),
            param(datetime64as, NDArrayDas),
            param(float, NDArrayF),
            param(int, NDArrayI),
            param(object, NDArrayO),
        ],
    )
    def test_dtype(self, dtype: Any, hint: Any) -> None:
        arr = empty(0, dtype=dtype)
        die_if_unbearable(arr, hint)

    @mark.parametrize(
        ("ndim", "hint"),
        [
            param(0, NDArray0),
            param(1, NDArray1),
            param(2, NDArray2),
            param(3, NDArray3),
        ],
    )
    def test_ndim(self, ndim: int, hint: Any) -> None:
        arr = empty(zeros(ndim, dtype=int), dtype=float)
        die_if_unbearable(arr, hint)

    @mark.parametrize(
        ("dtype", "ndim", "hint"),
        [
            # ndim 0
            param(bool, 0, NDArrayB0),
            param(datetime64D, 0, NDArrayD0),
            param(datetime64Y, 0, NDArrayDY0),
            param(datetime64M, 0, NDArrayDM0),
            param(datetime64W, 0, NDArrayDW0),
            param(datetime64D, 0, NDArrayDD0),
            param(datetime64h, 0, NDArrayDh0),
            param(datetime64m, 0, NDArrayDm0),
            param(datetime64s, 0, NDArrayDs0),
            param(datetime64ms, 0, NDArrayDms0),
            param(datetime64us, 0, NDArrayDus0),
            param(datetime64ns, 0, NDArrayDns0),
            param(datetime64ps, 0, NDArrayDps0),
            param(datetime64fs, 0, NDArrayDfs0),
            param(datetime64as, 0, NDArrayDas0),
            param(float, 0, NDArrayF0),
            param(int, 0, NDArrayI0),
            param(object, 0, NDArrayO0),
            # ndim 1
            param(bool, 1, NDArrayB1),
            param(datetime64D, 1, NDArrayD1),
            param(datetime64Y, 1, NDArrayDY1),
            param(datetime64M, 1, NDArrayDM1),
            param(datetime64W, 1, NDArrayDW1),
            param(datetime64D, 1, NDArrayDD1),
            param(datetime64h, 1, NDArrayDh1),
            param(datetime64m, 1, NDArrayDm1),
            param(datetime64s, 1, NDArrayDs1),
            param(datetime64ms, 1, NDArrayDms1),
            param(datetime64us, 1, NDArrayDus1),
            param(datetime64ns, 1, NDArrayDns1),
            param(datetime64ps, 1, NDArrayDps1),
            param(datetime64fs, 1, NDArrayDfs1),
            param(datetime64as, 1, NDArrayDas1),
            param(float, 1, NDArrayF1),
            param(int, 1, NDArrayI1),
            param(object, 1, NDArrayO1),
            # ndim 2
            param(bool, 2, NDArrayB2),
            param(datetime64D, 2, NDArrayD2),
            param(datetime64Y, 2, NDArrayDY2),
            param(datetime64M, 2, NDArrayDM2),
            param(datetime64W, 2, NDArrayDW2),
            param(datetime64D, 2, NDArrayDD2),
            param(datetime64h, 2, NDArrayDh2),
            param(datetime64m, 2, NDArrayDm2),
            param(datetime64s, 2, NDArrayDs2),
            param(datetime64ms, 2, NDArrayDms2),
            param(datetime64us, 2, NDArrayDus2),
            param(datetime64ns, 2, NDArrayDns2),
            param(datetime64ps, 2, NDArrayDps2),
            param(datetime64fs, 2, NDArrayDfs2),
            param(datetime64as, 2, NDArrayDas2),
            param(float, 2, NDArrayF2),
            param(int, 2, NDArrayI2),
            param(object, 2, NDArrayO2),
            # ndim 3
            param(bool, 3, NDArrayB3),
            param(datetime64D, 3, NDArrayD3),
            param(datetime64Y, 3, NDArrayDY3),
            param(datetime64M, 3, NDArrayDM3),
            param(datetime64W, 3, NDArrayDW3),
            param(datetime64D, 3, NDArrayDD3),
            param(datetime64h, 3, NDArrayDh3),
            param(datetime64m, 3, NDArrayDm3),
            param(datetime64s, 3, NDArrayDs3),
            param(datetime64ms, 3, NDArrayDms3),
            param(datetime64us, 3, NDArrayDus3),
            param(datetime64ns, 3, NDArrayDns3),
            param(datetime64ps, 3, NDArrayDps3),
            param(datetime64fs, 3, NDArrayDfs3),
            param(datetime64as, 3, NDArrayDas3),
            param(float, 3, NDArrayF3),
            param(int, 3, NDArrayI3),
            param(object, 3, NDArrayO3),
        ],
    )
    def test_compound(self, dtype: Any, ndim: int, hint: Any) -> None:
        arr = empty(zeros(ndim, dtype=int), dtype=dtype)
        die_if_unbearable(arr, hint)

    @given(arr=int_arrays())
    @example(arr=array([], dtype=int))
    @mark.parametrize("dtype", [param(int), param(float)])
    @mark.parametrize(
        "hint",
        [
            param(NDArrayINeg),
            param(NDArrayINonNeg),
            param(NDArrayINonPos),
            param(NDArrayINonZr),
            param(NDArrayIPos),
            param(NDArrayIZr),
            param(NDArrayI0Neg),
            param(NDArrayI0NonNeg),
            param(NDArrayI0NonPos),
            param(NDArrayI0NonZr),
            param(NDArrayI0Pos),
            param(NDArrayI0Zr),
            param(NDArrayI1Neg),
            param(NDArrayI1NonNeg),
            param(NDArrayI1NonPos),
            param(NDArrayI1NonZr),
            param(NDArrayI1Pos),
            param(NDArrayI1Zr),
            param(NDArrayI2Neg),
            param(NDArrayI2NonNeg),
            param(NDArrayI2NonPos),
            param(NDArrayI2NonZr),
            param(NDArrayI2Pos),
            param(NDArrayI2Zr),
            param(NDArrayI3Neg),
            param(NDArrayI3NonNeg),
            param(NDArrayI3NonPos),
            param(NDArrayI3NonZr),
            param(NDArrayI3Pos),
            param(NDArrayI3Zr),
        ],
    )
    @settings(max_examples=1, phases={Phase.explicit, Phase.generate})
    def test_int_checks(self, arr: NDArrayI, dtype: Any, hint: Any) -> None:
        with suppress(BeartypeDoorHintViolation):
            die_if_unbearable(arr.astype(dtype), hint)

    @given(arr=float_arrays())
    @example(arr=array([], dtype=float))
    @example(arr=array([nan], dtype=float))
    @example(arr=array([nan, nan], dtype=float))
    @mark.parametrize(
        "hint",
        [
            param(NDArrayFFin),
            param(NDArrayFFinInt),
            param(NDArrayFFinIntNan),
            param(NDArrayFFinNeg),
            param(NDArrayFFinNegNan),
            param(NDArrayFFinNonNeg),
            param(NDArrayFFinNonNegNan),
            param(NDArrayFFinNonPos),
            param(NDArrayFFinNonPosNan),
            param(NDArrayFFinNonZr),
            param(NDArrayFFinNonZrNan),
            param(NDArrayFFinPos),
            param(NDArrayFFinPosNan),
            param(NDArrayFFinNan),
            param(NDArrayFInt),
            param(NDArrayFIntNan),
            param(NDArrayFNeg),
            param(NDArrayFNegNan),
            param(NDArrayFNonNeg),
            param(NDArrayFNonNegNan),
            param(NDArrayFNonPos),
            param(NDArrayFNonPosNan),
            param(NDArrayFNonZr),
            param(NDArrayFNonZrNan),
            param(NDArrayFPos),
            param(NDArrayFPosNan),
            param(NDArrayFZr),
            param(NDArrayFZrNonMic),
            param(NDArrayFZrNonMicNan),
            param(NDArrayFZrNan),
            param(NDArrayFZrFinNonMic),
            param(NDArrayFZrFinNonMicNan),
            param(NDArrayF0Fin),
            param(NDArrayF0FinInt),
            param(NDArrayF0FinIntNan),
            param(NDArrayF0FinNeg),
            param(NDArrayF0FinNegNan),
            param(NDArrayF0FinNonNeg),
            param(NDArrayF0FinNonNegNan),
            param(NDArrayF0FinNonPos),
            param(NDArrayF0FinNonPosNan),
            param(NDArrayF0FinNonZr),
            param(NDArrayF0FinNonZrNan),
            param(NDArrayF0FinPos),
            param(NDArrayF0FinPosNan),
            param(NDArrayF0FinNan),
            param(NDArrayF0Int),
            param(NDArrayF0IntNan),
            param(NDArrayF0Neg),
            param(NDArrayF0NegNan),
            param(NDArrayF0NonNeg),
            param(NDArrayF0NonNegNan),
            param(NDArrayF0NonPos),
            param(NDArrayF0NonPosNan),
            param(NDArrayF0NonZr),
            param(NDArrayF0NonZrNan),
            param(NDArrayF0Pos),
            param(NDArrayF0PosNan),
            param(NDArrayF0Zr),
            param(NDArrayF0ZrNonMic),
            param(NDArrayF0ZrNonMicNan),
            param(NDArrayF0ZrNan),
            param(NDArrayF0ZrFinNonMic),
            param(NDArrayF0ZrFinNonMicNan),
            param(NDArrayF1Fin),
            param(NDArrayF1FinInt),
            param(NDArrayF1FinIntNan),
            param(NDArrayF1FinNeg),
            param(NDArrayF1FinNegNan),
            param(NDArrayF1FinNonNeg),
            param(NDArrayF1FinNonNegNan),
            param(NDArrayF1FinNonPos),
            param(NDArrayF1FinNonPosNan),
            param(NDArrayF1FinNonZr),
            param(NDArrayF1FinNonZrNan),
            param(NDArrayF1FinPos),
            param(NDArrayF1FinPosNan),
            param(NDArrayF1FinNan),
            param(NDArrayF1Int),
            param(NDArrayF1IntNan),
            param(NDArrayF1Neg),
            param(NDArrayF1NegNan),
            param(NDArrayF1NonNeg),
            param(NDArrayF1NonNegNan),
            param(NDArrayF1NonPos),
            param(NDArrayF1NonPosNan),
            param(NDArrayF1NonZr),
            param(NDArrayF1NonZrNan),
            param(NDArrayF1Pos),
            param(NDArrayF1PosNan),
            param(NDArrayF1Zr),
            param(NDArrayF1ZrNonMic),
            param(NDArrayF1ZrNonMicNan),
            param(NDArrayF1ZrNan),
            param(NDArrayF1ZrFinNonMic),
            param(NDArrayF1ZrFinNonMicNan),
            param(NDArrayF2Fin),
            param(NDArrayF2FinInt),
            param(NDArrayF2FinIntNan),
            param(NDArrayF2FinNeg),
            param(NDArrayF2FinNegNan),
            param(NDArrayF2FinNonNeg),
            param(NDArrayF2FinNonNegNan),
            param(NDArrayF2FinNonPos),
            param(NDArrayF2FinNonPosNan),
            param(NDArrayF2FinNonZr),
            param(NDArrayF2FinNonZrNan),
            param(NDArrayF2FinPos),
            param(NDArrayF2FinPosNan),
            param(NDArrayF2FinNan),
            param(NDArrayF2Int),
            param(NDArrayF2IntNan),
            param(NDArrayF2Neg),
            param(NDArrayF2NegNan),
            param(NDArrayF2NonNeg),
            param(NDArrayF2NonNegNan),
            param(NDArrayF2NonPos),
            param(NDArrayF2NonPosNan),
            param(NDArrayF2NonZr),
            param(NDArrayF2NonZrNan),
            param(NDArrayF2Pos),
            param(NDArrayF2PosNan),
            param(NDArrayF2Zr),
            param(NDArrayF2ZrNonMic),
            param(NDArrayF2ZrNonMicNan),
            param(NDArrayF2ZrNan),
            param(NDArrayF2ZrFinNonMic),
            param(NDArrayF2ZrFinNonMicNan),
            param(NDArrayF3Fin),
            param(NDArrayF3FinInt),
            param(NDArrayF3FinIntNan),
            param(NDArrayF3FinNeg),
            param(NDArrayF3FinNegNan),
            param(NDArrayF3FinNonNeg),
            param(NDArrayF3FinNonNegNan),
            param(NDArrayF3FinNonPos),
            param(NDArrayF3FinNonPosNan),
            param(NDArrayF3FinNonZr),
            param(NDArrayF3FinNonZrNan),
            param(NDArrayF3FinPos),
            param(NDArrayF3FinPosNan),
            param(NDArrayF3FinNan),
            param(NDArrayF3Int),
            param(NDArrayF3IntNan),
            param(NDArrayF3Neg),
            param(NDArrayF3NegNan),
            param(NDArrayF3NonNeg),
            param(NDArrayF3NonNegNan),
            param(NDArrayF3NonPos),
            param(NDArrayF3NonPosNan),
            param(NDArrayF3NonZr),
            param(NDArrayF3NonZrNan),
            param(NDArrayF3Pos),
            param(NDArrayF3PosNan),
            param(NDArrayF3Zr),
            param(NDArrayF3ZrNonMic),
            param(NDArrayF3ZrNonMicNan),
            param(NDArrayF3ZrNan),
            param(NDArrayF3ZrFinNonMic),
            param(NDArrayF3ZrFinNonMicNan),
        ],
    )
    @settings(max_examples=1, phases={Phase.explicit, Phase.generate})
    def test_float_checks(self, arr: NDArrayF, hint: Any) -> None:
        with suppress(BeartypeDoorHintViolation):
            die_if_unbearable(arr, hint)



from __future__ import annotations

from pathlib import Path

from nox import Session
from nox import session


@session
def ruff(session: Session) -> None:
    """Run `ruff`."""
    session.install("ruff")
    _ = session.run("ruff", "--fix", ".")


@session(python=["3.10", "3.11"])
def tests(session: Session, /) -> None:
    """Run the tests."""
    session.install("--upgrade", "pip-tools")
    requirements = set(Path(__file__).parent.glob("requirements*.txt"))
    _ = session.run("pip-sync", *(r.as_posix() for r in requirements))
    _ = session.run(
        "pytest", "--cov-report=term-missing:skip-covered", "-n=auto"
    )



from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING

from hypothesis import given
from hypothesis.errors import InvalidArgument
from hypothesis.strategies import DataObject
from hypothesis.strategies import booleans
from hypothesis.strategies import data
from hypothesis.strategies import floats
from hypothesis.strategies import integers
from hypothesis.strategies import none
from pandas.testing import assert_index_equal

from utilities.hypothesis import assume_does_not_raise
from utilities.hypothesis import bool_data_arrays
from utilities.hypothesis import dicts_of_indexes
from utilities.hypothesis import float_data_arrays
from utilities.hypothesis import int64s
from utilities.hypothesis import int_data_arrays
from utilities.hypothesis import str_data_arrays
from utilities.hypothesis import text_ascii
from utilities.hypothesis.xarray import _merge_into_dict_of_indexes

if TYPE_CHECKING:  # pragma: no cover
    from utilities.pandas import IndexA


class TestBoolDataArrays:
    @given(data=data(), indexes=dicts_of_indexes(), name=text_ascii() | none())
    def test_main(
        self, data: DataObject, indexes: Mapping[str, IndexA], name: str | None
    ) -> None:
        array = data.draw(bool_data_arrays(indexes, name=name))
        assert set(array.coords) == set(indexes)
        assert array.dims == tuple(indexes)
        assert array.dtype == bool
        assert array.name == name
        for arr, exp in zip(
            array.indexes.values(), indexes.values(), strict=True
        ):
            assert_index_equal(arr, exp, check_names=False)


class TestDictsOfIndexes:
    @given(
        data=data(),
        min_dims=integers(1, 3),
        max_dims=integers(1, 3) | none(),
        min_side=integers(1, 10),
        max_side=integers(1, 10) | none(),
    )
    def test_main(
        self,
        data: DataObject,
        min_dims: int,
        max_dims: int | None,
        min_side: int,
        max_side: int | None,
    ) -> None:
        with assume_does_not_raise(InvalidArgument):
            indexes = data.draw(
                dicts_of_indexes(
                    min_dims=min_dims,
                    max_dims=max_dims,
                    min_side=min_side,
                    max_side=max_side,
                )
            )
        ndims = len(indexes)
        assert ndims >= min_dims
        if max_dims is not None:
            assert ndims <= max_dims
        for index in indexes.values():
            length = len(index)
            assert length >= min_side
            if max_side is not None:
                assert length <= max_side


class TestFloatDataArrays:
    @given(
        data=data(),
        indexes=dicts_of_indexes(),
        min_value=floats() | none(),
        max_value=floats() | none(),
        allow_nan=booleans(),
        allow_inf=booleans(),
        allow_pos_inf=booleans(),
        allow_neg_inf=booleans(),
        integral=booleans(),
        unique=booleans(),
        name=text_ascii() | none(),
    )
    def test_main(
        self,
        *,
        data: DataObject,
        indexes: Mapping[str, IndexA],
        min_value: float | None,
        max_value: float | None,
        allow_nan: bool,
        allow_inf: bool,
        allow_pos_inf: bool,
        allow_neg_inf: bool,
        integral: bool,
        unique: bool,
        name: str | None,
    ) -> None:
        with assume_does_not_raise(InvalidArgument):
            array = data.draw(
                float_data_arrays(
                    indexes,
                    min_value=min_value,
                    max_value=max_value,
                    allow_nan=allow_nan,
                    allow_inf=allow_inf,
                    allow_pos_inf=allow_pos_inf,
                    allow_neg_inf=allow_neg_inf,
                    integral=integral,
                    unique=unique,
                    name=name,
                )
            )
        assert set(array.coords) == set(indexes)
        assert array.dims == tuple(indexes)
        assert array.dtype == float
        assert array.name == name
        for arr, exp in zip(
            array.indexes.values(), indexes.values(), strict=True
        ):
            assert_index_equal(arr, exp, check_names=False)


class TestIntDataArrays:
    @given(
        data=data(),
        indexes=dicts_of_indexes(),
        min_value=int64s() | none(),
        max_value=int64s() | none(),
        unique=booleans(),
        name=text_ascii() | none(),
    )
    def test_main(
        self,
        *,
        data: DataObject,
        indexes: Mapping[str, IndexA],
        min_value: int | None,
        max_value: int | None,
        unique: bool,
        name: str | None,
    ) -> None:
        with assume_does_not_raise(InvalidArgument):
            array = data.draw(
                int_data_arrays(
                    indexes,
                    min_value=min_value,
                    max_value=max_value,
                    unique=unique,
                    name=name,
                )
            )
        assert set(array.coords) == set(indexes)
        assert array.dims == tuple(indexes)
        assert array.dtype == int
        assert array.name == name
        for arr, exp in zip(
            array.indexes.values(), indexes.values(), strict=True
        ):
            assert_index_equal(arr, exp, check_names=False)


class TestMergeIntoDictOfIndexes:
    @given(data=data())
    def test_empty(self, data: DataObject) -> None:
        _ = data.draw(_merge_into_dict_of_indexes())

    @given(
        data=data(),
        indexes1=dicts_of_indexes() | none(),
        indexes2=dicts_of_indexes(),
    )
    def test_non_empty(
        self,
        data: DataObject,
        indexes1: Mapping[str, IndexA] | None,
        indexes2: Mapping[str, IndexA],
    ) -> None:
        indexes_ = data.draw(_merge_into_dict_of_indexes(indexes1, **indexes2))
        expected = (set() if indexes1 is None else set(indexes1)) | set(
            indexes2
        )
        assert set(indexes_) == expected


class TestStrDataArrays:
    @given(
        data=data(),
        indexes=dicts_of_indexes(),
        min_size=integers(0, 100),
        max_size=integers(0, 100) | none(),
        allow_none=booleans(),
        unique=booleans(),
        name=text_ascii() | none(),
    )
    def test_main(
        self,
        *,
        data: DataObject,
        indexes: Mapping[str, IndexA],
        min_size: int,
        max_size: int | None,
        allow_none: bool,
        unique: bool,
        name: str | None,
    ) -> None:
        with assume_does_not_raise(InvalidArgument):
            array = data.draw(
                str_data_arrays(
                    indexes,
                    min_size=min_size,
                    max_size=max_size,
                    allow_none=allow_none,
                    unique=unique,
                    name=name,
                )
            )
        assert set(array.coords) == set(indexes)
        assert array.dims == tuple(indexes)
        assert array.dtype == object
        assert array.name == name
        for arr, exp in zip(
            array.indexes.values(), indexes.values(), strict=True
        ):
            assert_index_equal(arr, exp, check_names=False)



from __future__ import annotations

import datetime as dt
from collections.abc import Sized
from typing import Any
from typing import cast

import sqlalchemy
from hypothesis import given
from hypothesis.strategies import DataObject
from hypothesis.strategies import SearchStrategy
from hypothesis.strategies import booleans
from hypothesis.strategies import data
from hypothesis.strategies import floats
from hypothesis.strategies import integers
from hypothesis.strategies import just
from hypothesis.strategies import lists
from hypothesis.strategies import none
from hypothesis_sqlalchemy.sample import table_records_lists
from pandas import DataFrame
from pandas import NaT
from pandas import Series
from pytest import mark
from pytest import param
from pytest import raises
from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import Date
from sqlalchemy import DateTime
from sqlalchemy import Engine
from sqlalchemy import Float
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import String
from sqlalchemy import Table
from sqlalchemy import func
from sqlalchemy import insert
from sqlalchemy import select
from sqlalchemy.exc import DuplicateColumnError
from sqlalchemy.orm import declarative_base

from utilities.hypothesis import dates_pd
from utilities.hypothesis import datetimes_pd
from utilities.hypothesis import int64s
from utilities.hypothesis import sqlite_engines
from utilities.hypothesis import text_ascii
from utilities.numpy import datetime64ns
from utilities.pandas import Int64
from utilities.pandas import boolean
from utilities.pandas import datetime64nsutc
from utilities.pandas import string
from utilities.sqlalchemy import DatesWithTimeComponentsError
from utilities.sqlalchemy import NonPositiveStreamError
from utilities.sqlalchemy import SeriesAgainstTableColumnError
from utilities.sqlalchemy import SeriesNameNotInTableError
from utilities.sqlalchemy import SeriesNameSnakeCaseNotInTableError
from utilities.sqlalchemy import ensure_table_created
from utilities.sqlalchemy import get_table
from utilities.sqlalchemy import insert_dataframe
from utilities.sqlalchemy import insert_items
from utilities.sqlalchemy import select_to_dataframe
from utilities.sqlalchemy.pandas import _check_select_for_duplicates
from utilities.sqlalchemy.pandas import _check_series_against_table_column
from utilities.sqlalchemy.pandas import _dataframe_columns_to_snake
from utilities.sqlalchemy.pandas import _parse_series_against_table
from utilities.sqlalchemy.pandas import _rows_to_dataframe
from utilities.sqlalchemy.pandas import _stream_dataframes
from utilities.sqlalchemy.pandas import _table_column_to_dtype
from utilities.sqlalchemy.pandas import _yield_dataframe_rows_as_dicts
from utilities.sqlalchemy.pandas import _yield_insertion_elements
from utilities.text import snake_case
from utilities.types import NoneType


class TestCheckSelectForDuplicates:
    def test_error(self) -> None:
        table = Table(
            "example", MetaData(), Column("id", Integer, primary_key=True)
        )
        sel = select(table.c.id, table.c.id)
        with raises(DuplicateColumnError):
            _check_select_for_duplicates(sel)


class TestCheckSeriesAgainstTableColumn:
    @mark.parametrize(
        ("dtype", "column_type"),
        [
            param(bool, Boolean),
            param(boolean, Boolean),
            param(bool, Integer),
            param(boolean, Integer),
            param(float, Float),
            param(datetime64ns, Date),
            param(datetime64nsutc, DateTime),
            param(int, Integer),
            param(Int64, Integer),
            param(string, String),
        ],
    )
    def test_success(self, dtype: Any, column_type: Any) -> None:
        series = Series([], dtype=dtype)
        column = Column("id", column_type)
        _check_series_against_table_column(series, column)

    @mark.parametrize(
        ("dtype", "column_type"),
        [param(object, Integer), param(datetime64ns, DateTime)],
    )
    def test_error(self, dtype: Any, column_type: Any) -> None:
        series = Series([], dtype=dtype)
        column = Column("id", column_type)
        with raises(SeriesAgainstTableColumnError):
            _check_series_against_table_column(series, column)

    def test_allow_naive_datetimes(self) -> None:
        series = Series([], dtype=datetime64ns)
        column = Column("id", DateTime)
        _check_series_against_table_column(
            series, column, allow_naive_datetimes=True
        )


class TestDataFrameColumnsToSnake:
    @given(col_name=text_ascii())
    def test_main(self, col_name: str) -> None:
        df = DataFrame(columns=[col_name])
        snake = _dataframe_columns_to_snake(df)
        assert snake.columns.tolist() == [snake_case(col_name)]


class TestInsertDataFrame:
    @given(data=data(), engine=sqlite_engines())
    def test_main(self, data: DataObject, engine: Engine) -> None:
        table = Table(
            "example", MetaData(), Column("id", Integer, primary_key=True)
        )
        ensure_table_created(table, engine)
        rows = data.draw(table_records_lists(table, max_size=10))
        df = DataFrame(rows, columns=["id"]).astype(int)
        insert_dataframe(df, table, engine)
        with engine.begin() as conn:
            res = conn.execute(select(func.count()).select_from(table)).scalar()
        assert res == len(rows)


class TestInsertItems:
    @given(data=data(), engine=sqlite_engines())
    def test_lists_of_tuples(self, data: DataObject, engine: Engine) -> None:
        table = Table(
            "example", MetaData(), Column("id", Integer, primary_key=True)
        )
        ensure_table_created(table, engine)
        rows = data.draw(table_records_lists(table, max_size=10))
        insert_items([(rows, table)], engine)
        with engine.begin() as conn:
            res = conn.execute(select(func.count()).select_from(table)).scalar()
        assert res == len(rows)

    @given(data=data(), engine=sqlite_engines())
    def test_dataframe_with_table(
        self, data: DataObject, engine: Engine
    ) -> None:
        table = Table(
            "example", MetaData(), Column("id", Integer, primary_key=True)
        )
        ensure_table_created(table, engine)
        rows = data.draw(table_records_lists(table, max_size=10))
        df = DataFrame(rows, columns=["id"]).astype(int)
        insert_items([(df, table)], engine)
        with engine.begin() as conn:
            res = conn.execute(select(func.count()).select_from(table)).scalar()
        assert res == len(rows)

    @given(data=data(), engine=sqlite_engines())
    def test_model(self, data: DataObject, engine: Engine) -> None:
        class Example(declarative_base()):
            __tablename__ = "example"

            id_ = Column(Integer, primary_key=True)

        ensure_table_created(Example, engine)
        rows = data.draw(table_records_lists(get_table(Example), max_size=10))
        items = [Example(id_=id_) for id_, in rows]
        insert_items(items, engine)
        with engine.begin() as conn:
            res = conn.execute(
                select(func.count()).select_from(get_table(Example))
            ).scalar()
        assert res == len(rows)

    @given(engine=sqlite_engines())
    def test_type_error(self, engine: Engine) -> None:
        table = Table(
            "example", MetaData(), Column("id", Integer, primary_key=True)
        )
        items = [(None, table)]
        with raises(TypeError, match="Invalid type: first="):
            insert_items(items, engine)


class TestParseSeriesAgainstTable:
    @mark.parametrize(
        ("series_name", "table_column_name", "snake"),
        [
            param("id", "id", False),
            param("id", "Id", True),
            param("Id", "id", True),
        ],
    )
    def test_main(
        self, *, series_name: str, table_column_name: str, snake: bool
    ) -> None:
        series = Series([], dtype=int, name=series_name)
        table = Table(
            "example",
            MetaData(),
            Column(table_column_name, Integer, primary_key=True),
        )
        key, _ = _parse_series_against_table(series, table, snake=snake)
        assert key == table_column_name

    @mark.parametrize(
        ("snake", "error"),
        [
            param(True, SeriesNameSnakeCaseNotInTableError),
            param(False, SeriesNameNotInTableError),
        ],
    )
    def test_error(self, *, snake: bool, error: type[Exception]) -> None:
        series = Series([], dtype=int, name="name")
        table = Table(
            "example", MetaData(), Column("id", Integer, primary_key=True)
        )
        with raises(error):
            _ = _parse_series_against_table(series, table, snake=snake)


class TestRowsToDataFrame:
    @given(data=data(), engine=sqlite_engines())
    @mark.parametrize(
        ("col_name", "snake"), [param("id", False), param("Id", True)]
    )
    def test_main(
        self, *, data: DataObject, col_name: str, engine: Engine, snake: bool
    ) -> None:
        table = Table(
            "example", MetaData(), Column(col_name, Integer, primary_key=True)
        )
        ensure_table_created(table, engine)
        values = data.draw(table_records_lists(table, min_size=1, max_size=10))
        with engine.begin() as conn:
            _ = conn.execute(insert(table).values(values))
        with engine.begin() as conn:
            rows = conn.execute(sel := select(table)).all()
        df = _rows_to_dataframe(sel, rows, snake=snake)
        assert len(df) == len(values)
        assert dict(df.dtypes) == {"id": Int64}


class TestSelectToDataFrame:
    @given(
        data=data(), engine=sqlite_engines(), stream=integers(1, 10) | none()
    )
    def test_main(
        self, data: DataObject, engine: Engine, stream: int | None
    ) -> None:
        table = Table(
            "example", MetaData(), Column("id", Integer, primary_key=True)
        )
        ensure_table_created(table, engine)
        rows = data.draw(table_records_lists(table, min_size=1, max_size=10))
        insert_items([(rows, table)], engine)
        result = select_to_dataframe(select(table), engine, stream=stream)
        if stream is None:
            assert isinstance(result, DataFrame)
            assert len(cast(Sized, result)) == len(rows)
            assert dict(cast(Any, result).dtypes) == {"id": Int64}
        else:
            assert not isinstance(result, DataFrame)
            for df in result:
                assert 1 <= len(df) <= stream
                assert dict(df.dtypes) == {"id": Int64}


class TestStreamDataFrames:
    @given(data=data(), engine=sqlite_engines(), stream=integers(1, 10))
    def test_main(self, data: DataObject, engine: Engine, stream: int) -> None:
        table = Table(
            "example", MetaData(), Column("id", Integer, primary_key=True)
        )
        ensure_table_created(table, engine)
        rows = data.draw(table_records_lists(table, min_size=1, max_size=10))
        insert_items([(rows, table)], engine)
        for df in _stream_dataframes(select(table), engine, stream):
            assert 1 <= len(df) <= stream
            assert dict(df.dtypes) == {"id": Int64}

    @given(engine=sqlite_engines())
    def test_non_positive_stream(self, engine: Engine) -> None:
        table = Table(
            "example", MetaData(), Column("id", Integer, primary_key=True)
        )
        with raises(NonPositiveStreamError):
            _ = list(_stream_dataframes(select(table), engine, 0))


class TestTableColumnToDtype:
    @mark.parametrize(
        ("column", "expected"),
        [
            param(Column(Boolean), boolean),
            param(Column(Date), datetime64ns),
            param(Column(DateTime), datetime64ns),
            param(Column(Float), float),
            param(Column(Integer), Int64),
            param(Column(String), string),
            param(Column(sqlalchemy.DECIMAL), float),
        ],
    )
    def test_main(self, column: Any, expected: Any) -> None:
        assert _table_column_to_dtype(column) == expected


class TestYieldDataFrameRowsAsDicts:
    @given(data=data())
    @mark.parametrize(
        ("elements", "df_dtype", "table_dtype", "result_type"),
        [
            param(booleans(), bool, Boolean, bool),
            param(booleans() | none(), boolean, Boolean, (bool, NoneType)),
            param(dates_pd(), datetime64ns, Date, dt.date),
            param(
                dates_pd() | just(NaT) | none(),
                datetime64ns,
                Date,
                (dt.date, NoneType),
            ),
            param(datetimes_pd(), datetime64nsutc, DateTime, dt.datetime),
            param(
                datetimes_pd() | just(NaT) | none(),
                datetime64nsutc,
                DateTime,
                (dt.datetime, NoneType),
            ),
            param(floats(), float, Float, (float, NoneType)),
            param(int64s(), int, Integer, int),
            param(int64s() | none(), Int64, Integer, (int, NoneType)),
            param(text_ascii(), string, String, str),
            param(text_ascii() | none(), string, String, (str, NoneType)),
        ],
    )
    def test_main(
        self,
        data: DataObject,
        elements: SearchStrategy[Any],
        df_dtype: Any,
        table_dtype: Any,
        result_type: Any,
    ) -> None:
        values = data.draw(lists(elements, max_size=10))
        df = DataFrame(values, columns=["id"], dtype=df_dtype)
        table = Table(
            "example", MetaData(), Column("id", table_dtype, primary_key=True)
        )
        dicts = list(_yield_dataframe_rows_as_dicts(df, table))
        assert len(dicts) == len(values)
        for dict_ in dicts:
            assert isinstance(dict_["id"], result_type)


class TestYieldInsertionElements:
    @given(data=data())
    @mark.parametrize(
        ("elements", "dtype", "expected"),
        [
            param(booleans(), bool, bool),
            param(booleans() | none(), boolean, (bool, NoneType)),
            param(dates_pd(), datetime64ns, dt.date),
            param(
                dates_pd() | just(NaT) | none(),
                datetime64ns,
                (dt.date, NoneType),
            ),
            param(datetimes_pd(), datetime64nsutc, dt.datetime),
            param(
                datetimes_pd() | just(NaT) | none(),
                datetime64nsutc,
                (dt.datetime, NoneType),
            ),
            param(floats(), float, (float, NoneType)),
            param(int64s(), int, int),
            param(int64s() | none(), Int64, (int, NoneType)),
            param(text_ascii(), string, str),
            param(text_ascii() | none(), string, (str, NoneType)),
        ],
    )
    def test_main(
        self,
        data: DataObject,
        elements: SearchStrategy[Any],
        dtype: Any,
        expected: type | tuple[type, ...],
    ) -> None:
        values = data.draw(lists(elements, max_size=10))
        series = Series(values, dtype=dtype)
        for el in _yield_insertion_elements(series):
            assert isinstance(el, expected)

    @given(date=dates_pd())
    def test_dates_with_time_components_error(self, date: dt.date) -> None:
        series = Series(
            [dt.datetime.combine(date, dt.time(12))], dtype=datetime64ns
        )
        with raises(DatesWithTimeComponentsError):
            _ = list(_yield_insertion_elements(series))

    def test_type_error(self) -> None:
        series = Series(dtype=object)
        with raises(TypeError, match="Invalid dtype"):
            _ = list(_yield_insertion_elements(series))



from __future__ import annotations

from .holoviews import apply_opts
from .holoviews import relabel_plot
from .holoviews import save_plot

__all__ = ["apply_opts", "relabel_plot", "save_plot"]


try:
    from .xarray import ArrayNameIsEmptyStringError
    from .xarray import ArrayNameNotAStringError
    from .xarray import plot_curve
except ModuleNotFoundError:  # pragma: no cover
    pass
else:
    __all__ += [
        "ArrayNameIsEmptyStringError",
        "ArrayNameNotAStringError",
        "plot_curve",
    ]



from __future__ import annotations

from click import command
from click import echo
from click.testing import CliRunner
from hypothesis import given
from hypothesis.strategies import integers
from hypothesis.strategies import none
from pytest import mark
from pytest import param

from utilities.click import local_scheduler_option_default_central
from utilities.click import local_scheduler_option_default_local
from utilities.click import workers_option
from utilities.typing import SequenceStrs


class TestLocalSchedulerOption:
    @mark.parametrize(
        ("args", "expected"),
        [param([], True), param(["-ls"], True), param(["-nls"], False)],
    )
    def test_default_local(self, *, args: SequenceStrs, expected: bool) -> None:
        @command()
        @local_scheduler_option_default_local
        def cli(*, local_scheduler: bool) -> None:
            echo(f"local_scheduler = {local_scheduler}")

        result = CliRunner().invoke(cli, args)
        assert result.exit_code == 0
        assert result.stdout == f"local_scheduler = {expected}\n"

    @mark.parametrize(
        ("args", "expected"),
        [param([], False), param(["-ls"], True), param(["-nls"], False)],
    )
    def test_default_central(
        self, *, args: SequenceStrs, expected: bool
    ) -> None:
        @command()
        @local_scheduler_option_default_central
        def cli(*, local_scheduler: bool) -> None:
            echo(f"local_scheduler = {local_scheduler}")

        result = CliRunner().invoke(cli, args)
        assert result.exit_code == 0
        assert result.stdout == f"local_scheduler = {expected}\n"


class TestWorkersOption:
    @given(workers=integers() | none())
    def test_main(self, workers: int | None) -> None:
        @command()
        @workers_option
        def cli(*, workers: int | None) -> None:
            echo(f"workers = {workers}")

        args = [] if workers is None else ["--workers", str(workers)]
        result = CliRunner().invoke(cli, args)
        assert result.exit_code == 0
        assert result.stdout == f"workers = {workers}\n"



from __future__ import annotations

from pytest import mark
from pytest import param
from pytest import raises

from utilities.re import MultipleCaptureGroupsError
from utilities.re import MultipleMatchesError
from utilities.re import NoCaptureGroupsError
from utilities.re import NoMatchesError
from utilities.re import extract_group
from utilities.re import extract_groups
from utilities.typing import IterableStrs


class TestExtractGroup:
    def test_success(self) -> None:
        assert extract_group(r"(\d)", "A0A") == "0"

    def test_no_groups(self) -> None:
        with raises(NoCaptureGroupsError):
            _ = extract_group(r"\d", "0")

    def test_multiple_groups(self) -> None:
        with raises(MultipleCaptureGroupsError):
            _ = extract_group(r"(\d)(\w)", "0A")

    def test_no_matches(self) -> None:
        with raises(NoMatchesError, match="pattern='.*', text='.*'"):
            _ = extract_group(r"(\d)", "A")

    def test_multiple_matches(self) -> None:
        with raises(MultipleMatchesError, match="pattern='.*', text='.*'"):
            _ = extract_group(r"(\d)", "0A0")


class TestExtractGroups:
    @mark.parametrize(
        ("pattern", "text", "expected"),
        [param(r"(\d)", "A0A", ["0"]), param(r"(\d)(\w)", "A0A0", ["0", "A"])],
    )
    def test_success(
        self, pattern: str, text: str, expected: IterableStrs
    ) -> None:
        assert extract_groups(pattern, text) == expected

    def test_no_groups(self) -> None:
        with raises(NoCaptureGroupsError):
            _ = extract_groups(r"\d", "0")

    @mark.parametrize(
        ("pattern", "text"), [param(r"(\d)", "A"), param(r"(\d)(\w)", "A0")]
    )
    def test_no_matches(self, pattern: str, text: str) -> None:
        with raises(NoMatchesError, match="pattern='.*', text='.*'"):
            _ = extract_groups(pattern, text)

    @mark.parametrize(
        ("pattern", "text"), [param(r"(\d)", "0A0"), param(r"(\d)(\w)", "0A0A")]
    )
    def test_multiple_matches(self, pattern: str, text: str) -> None:
        with raises(MultipleMatchesError, match="pattern='.*', text='.*'"):
            _ = extract_groups(pattern, text)



from __future__ import annotations

from .xarray import DataArray0
from .xarray import DataArray1
from .xarray import DataArray2
from .xarray import DataArray3
from .xarray import DataArrayB
from .xarray import DataArrayB0
from .xarray import DataArrayB1
from .xarray import DataArrayB2
from .xarray import DataArrayB3
from .xarray import DataArrayDns
from .xarray import DataArrayDns0
from .xarray import DataArrayDns1
from .xarray import DataArrayDns2
from .xarray import DataArrayDns3
from .xarray import DataArrayF
from .xarray import DataArrayF0
from .xarray import DataArrayF1
from .xarray import DataArrayF2
from .xarray import DataArrayF3
from .xarray import DataArrayI
from .xarray import DataArrayI0
from .xarray import DataArrayI1
from .xarray import DataArrayI2
from .xarray import DataArrayI3
from .xarray import DataArrayO
from .xarray import DataArrayO0
from .xarray import DataArrayO1
from .xarray import DataArrayO2
from .xarray import DataArrayO3

__all__ = [
    "DataArray0",
    "DataArray1",
    "DataArray2",
    "DataArray3",
    "DataArrayB",
    "DataArrayB0",
    "DataArrayB1",
    "DataArrayB2",
    "DataArrayB3",
    "DataArrayDns",
    "DataArrayDns0",
    "DataArrayDns1",
    "DataArrayDns2",
    "DataArrayDns3",
    "DataArrayF",
    "DataArrayF0",
    "DataArrayF1",
    "DataArrayF2",
    "DataArrayF3",
    "DataArrayI",
    "DataArrayI0",
    "DataArrayI1",
    "DataArrayI2",
    "DataArrayI3",
    "DataArrayO",
    "DataArrayO0",
    "DataArrayO1",
    "DataArrayO2",
    "DataArrayO3",
]


try:
    from .numbagg import ewma
    from .numbagg import exp_moving_sum
except ModuleNotFoundError:  # pragma: no cover
    pass
else:
    __all__ += ["ewma", "exp_moving_sum"]



from __future__ import annotations

from collections.abc import Sequence
from itertools import chain
from typing import Any

from hypothesis import given
from hypothesis.strategies import DataObject
from hypothesis.strategies import data
from hypothesis.strategies import integers
from hypothesis.strategies import lists
from hypothesis.strategies import sampled_from
from hypothesis.strategies import sets
from pytest import mark
from pytest import param
from pytest import raises

from utilities.iterables import IterableContainsDuplicatesError
from utilities.iterables import check_duplicates
from utilities.iterables import is_iterable_not_str


class TestCheckDuplicates:
    @given(x=sets(integers()))
    def test_main(self, x: set[int]) -> None:
        check_duplicates(x)

    @given(data=data(), x=lists(integers(), min_size=1))
    def test_error(self, data: DataObject, x: Sequence[int]) -> None:
        x_i = data.draw(sampled_from(x))
        y = chain(x, [x_i])
        with raises(IterableContainsDuplicatesError):
            check_duplicates(y)


class TestIsIterableNotStr:
    @mark.parametrize(
        ("x", "expected"),
        [
            param(None, False),
            param([], True),
            param((), True),
            param("", False),
        ],
    )
    def test_main(self, *, x: Any, expected: bool) -> None:
        assert is_iterable_not_str(x) is expected



from __future__ import annotations

from pathlib import Path

from hypothesis import given
from hypothesis.strategies import DataObject
from hypothesis.strategies import data
from pytest import raises

from utilities.git import InvalidRepoError
from utilities.git import get_branch_name
from utilities.git import get_repo_name
from utilities.git import get_repo_root
from utilities.hypothesis import git_repos
from utilities.hypothesis import text_ascii
from utilities.tempfile import TemporaryDirectory


class TestGetBranchName:
    @given(data=data(), branch=text_ascii(min_size=1))
    def test_main(self, *, data: DataObject, branch: str) -> None:
        repo = data.draw(git_repos(branch=branch))
        result = get_branch_name(cwd=repo.path)
        assert result == branch


class TestGetRepoName:
    def test_main(self) -> None:
        result = get_repo_name()
        expected = "python-utilities"
        assert result == expected


class TestGetRepoRoot:
    @given(repo=git_repos())
    def test_main(self, *, repo: TemporaryDirectory) -> None:
        path = repo.path
        result = get_repo_root(cwd=path)
        assert result == path.resolve()

    def test_error(self, *, tmp_path: Path) -> None:
        with raises(InvalidRepoError):
            _ = get_repo_root(cwd=tmp_path)
